/* A taglet to be used in describing the ports of the actors.

 Copyright (c) 2003-2004 The Regents of the University of California.
 All rights reserved.
 Permission is hereby granted, without written agreement and without
 license or royalty fees, to use, copy, modify, and distribute this
 software and its documentation for any purpose, provided that the above
 copyright notice and the following two paragraphs appear in all copies
 of this software.

 IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 SUCH DAMAGE.

 THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 ENHANCEMENTS, OR MODIFICATIONS.

 PT_COPYRIGHT_VERSION_2
 COPYRIGHTENDKEY

 */

package doc.taglets;

import com.sun.tools.doclets.Taglet;
import com.sun.javadoc.*;
import java.util.Map;

/**
 * A sample Taglet representing @port.description. This tag can be used in any 
 * kind of {@link com.sun.javadoc.Doc}.  It is not an inline tag. The text is 
 * displayed in light blue to remind the developers to use this tag. ;)  For
 * example, "@port.description portName This is an input port that accepts some
 * value." would be shown as:
 * <DL>
 * <DT>
 * <B>:</B>
 * <DD><table cellpadding=2 cellspacing=0><tr><td bgcolor="light blue">
 *   web service
 * </td></tr></table></DD>
 * </DL>
 * If there are more than one tags like: "@category remote",
 * "@category distributed", they need to be tagged seperately.
 *
 * @author Ilkay Altintas
 * @version $Id: DescriptionTaglet.java,v 1.1 2005/03/03 17:32:17 altintas Exp $
 * @since Kepler 1.0
 */

public class DescriptionTaglet implements Taglet {

    private static final String NAME = "entity.description";
    private static final String HEADER = "Description:";

    /**
     * Return the name of this custom tag.
     */
    public String getName() {
        return NAME;
    }

    /**
     * Will return false since <code>@category</code>
     * cannot be used in field documentation.
     */
    public boolean inField() {
        return true;
    }

    /**
     * Will return false since <code>@category</code>
     * cannot be used in constructor documentation.
     */
    public boolean inConstructor() {
        return false;
    }

    /**
     * Will return false since <code>@category</code>
     * cannot be used in method documentation.
     */
    public boolean inMethod() {
        return false;
    }

    /**
     * Will return true since <code>@category</code>
     * can be used in method documentation.
     */
    public boolean inOverview() {
        return false;
    }

    /**
     * Will return true since <code>@category</code>
     * can be used in package documentation.
     */
    public boolean inPackage() {
        return false;
    }

    /**
     * Will return true since <code>@category</code>
     * can be used in type documentation (classes or interfaces).
     */
    public boolean inType() {
        return true;
    }

    /**
     * Will return false since <code>@category</code>
     * is not an inline tag.
     */

    public boolean isInlineTag() {
        return false;
    }

    /**
     * Register this Taglet.
     * @param tagletMap  the map to register this tag to.
     */
    public static void register(Map tagletMap) {
       DescriptionTaglet tag = new DescriptionTaglet();
       Taglet t = (Taglet) tagletMap.get(tag.getName());
       if (t != null) {
           tagletMap.remove(tag.getName());
       }
       tagletMap.put(tag.getName(), tag);
    }

    /**
     * Given the <code>Tag</code> representation of this custom
     * tag, return its string representation.
     * @param tag   the <code>Tag</code> representation of this custom tag.
     */
    public String toString(Tag tag) {
        return "<DT><B>" + HEADER + "</B><DD>"
               + "<table cellpadding=2 cellspacing=0><tr><td bgcolor=\"#FFFFCC\">"
               + tag.text()
               + "</td></tr></table></DD>\n";
    }

    /**
     * Given an array of <code>Tag</code>s representing this custom
     * tag, return its string representation.
     * @param tags  the array of <code>Tag</code>s representing of this custom tag.
     */
    public String toString(Tag[] tags) {
        if (tags.length == 0) {
            return null;
        }
        if(! (tags[0].holder() instanceof FieldDoc)) {
            return "";
        }
        String result = "\n<DT><B>" + HEADER + "</B><DD>";
        result += "<table cellpadding=2 cellspacing=0><tr><td bgcolor=\"#FFFFCC\">";
        for (int i = 0; i < tags.length; i++) {
            if (i > 0) {
                result += ", ";
            }
            result += tags[i].text();
        }
        return result + "</td></tr></table></DD>\n";
    }

    public String[] getTagString(Tag tag) {
        String result = tag.text().trim();
        String delimiter = " ";
        if(result.startsWith("\"")) {
            delimiter = "\"";
        }

        int pos = result.indexOf(delimiter, 1);
        if(-1 == pos) {
            throw new RuntimeException("Invalid " + this.getName() + " tag: '" + result + "'.");
        }

        String key = result.substring(0, pos).trim();
        if(key.startsWith("\"")) {
            key = key.substring(1);
        }
        if(key.endsWith("\"")) {
            key = key.substring(0, key.length() - 1);
        }

        String value = result.substring(pos + 1).trim();

        return new String[] {key, value};
    }
}
