package edu.stanford.db.rdf.model.inv;

import java.util.*;
import org.w3c.rdf.model.*;
import com.lastmileservices.rdf.util.Cache;

public class InverterImpl implements Inverter {

  // Resource -> InvertedResource
  Cache imap;

  static Inverter defaultCacheInverter = new InverterImpl(true);
  static Inverter defaultNoCacheInverter = new InverterImpl(false);

  private InverterImpl(boolean cache) {

    if(cache)
      imap = new Cache(0, false);
  }

  public static Inverter createInverter(boolean cache) {

    return cache ? defaultCacheInverter : defaultNoCacheInverter;
  }

  public Model enableInverted(Model m) {

    if(m instanceof InvertedModel)
      return m;
    return new InvertedModel(m);
  }

  public Model disableInverted(Model m) {

    if(m instanceof InvertedModel)
      return ((InvertedModel)m).getNonInverted();
    return m;
  }

  public boolean isEnabledInverted(Model m) {

    return (m instanceof InvertedModel);
  }

  /**
   * inverts a resource: already inverted becomes not inverted
   */
  public Resource invert(Resource r) {

    if(r instanceof InvertedResource)
      return ((InvertedResource)r).getNonInverted();
    if(r == null)
      return null;

    if(imap != null) {

      InvertedResource inv = (InvertedResource)imap.get(r);
      if(inv == null)
	imap.put(r, inv = new InvertedResourceImpl(r));
      return inv;

    } else
      return new InvertedResourceImpl(r);
  }

  public boolean isInverted(Resource r) {

    return (r instanceof InvertedResource);
  }

  public Resource getNonInverted(Resource r) {

    return isInverted(r) ? invert(r) : r;
  }

  static Resource makeResource(RDFNode o) {

    if(o instanceof Resource)
      return (Resource)o;
    if(o == null)
      return null;
    return new ResourceWrapper(o);
  }


  static class InvertedResourceImpl implements InvertedResource {

    static final String NS = "inverted:";

    Resource peer;

    public InvertedResourceImpl(Resource peer) {

      this.peer = peer;
    }

    public int hashCode() {

      return peer.hashCode() * 37;
    }

    public Resource getNonInverted() {

      return peer;
    }

    public boolean equals(Object that) {

      if(this == that)
	return true;
      if(!(that instanceof InvertedResource))
	return false;

      InvertedResource r = (InvertedResource)that;
      return peer.equals(r.getNonInverted());
    }

    public String toString() {

      try {
	return getURI();
      } catch (ModelException any) {
	return null;
      }
    }

    public String getURI() throws ModelException {

      return NS + peer.getURI();
    }

    public String getLabel() throws ModelException {

      return getURI();
    }

    public String getNamespace() {

      return null;
    }

    public String getLocalName() throws ModelException {

      return getURI();
    }
  }

  class InvertedStatement extends InvertedResourceImpl implements Statement {

    public InvertedStatement(Statement peer) {

      super(peer);
      //      System.err.println("Created inverted statement " + this + " of " + peer);
    }

    public Resource subject() throws ModelException {

      Statement st = (Statement)peer;
      return makeResource(st.object());
    }

    public Resource predicate() throws ModelException {

      Statement st = (Statement)peer;
      return invert(st.predicate());
    }

    public RDFNode object() throws ModelException {

      Statement st = (Statement)peer;
      return st.subject();
    }

    public String toString() {

      try {
	return "inverted(" + subject() + "," + predicate() + "," + object() + ")";
      } catch (ModelException any) {
	return null;
      }
    }
  }

  static class ResourceWrapper implements Resource {

    RDFNode peer;

    public ResourceWrapper(RDFNode n) {

      this.peer = n;
    }

    public String getLabel() throws ModelException {

      return peer.getLabel();
    }

    public String getURI() throws ModelException  {

      return getLabel();
    }

    public String getNamespace() {

      return null;
    }

    public String getLocalName() throws ModelException  {

      return getLabel();
    }

    public boolean equals(Object that) {

      return peer.equals(that);
    }

    public int hashCode() {

      return peer.hashCode();
    }
  }

  class InvertedModel implements Model {

    Model peer;

    public InvertedModel(Model m) {

      this.peer = m;
    }

    public Model getNonInverted() {

      return peer;
    }

    public void setSourceURI(String uri) throws ModelException {

      peer.setSourceURI(uri);
    }

    public String getSourceURI() throws ModelException {

      return peer.getSourceURI();
    }

    public int size() throws ModelException {

      return peer.size() * 2;
    }

    public boolean isEmpty() throws ModelException {

      return peer.isEmpty();
    }

    public String getLabel() throws ModelException {

      return peer.getLabel();
    }

    public String getURI() throws ModelException {

      return peer.getURI();
    }

    public String getNamespace() throws ModelException {

      return peer.getNamespace();
    }

    public String getLocalName() throws ModelException {

      return peer.getLocalName();
    }

    public Model find(Resource s, Resource p, RDFNode o) throws ModelException {

      Model res1, res2;

      if(p == null) {

	res1 = peer.find(s, null, o);
	res2 = peer.find(makeResource(o), null, s);

      } else if(!isInverted(p)) {

	res1 = peer.find(s, p, o);
	res2 = null;

      } else {

	res2 = peer.find(makeResource(o), isInverted(p) ? invert(p) : p, s);
	res1 = res2.create();
      }

      // find in both directions
//       Resource straight = isInverted(p) ? invert(p) : p;

//       Model res1 = peer.find(s, straight, o);
      
//       //      if(straight != null) {
//       Model res2 = peer.find(makeResource(o), invert(straight), s);
      //      NodeFactory f = res2.getNodeFactory();

      if(res2 != null) {
	for(Enumeration en = res2.elements(); en.hasMoreElements();)
	  res1.add(new InvertedStatement((Statement)en.nextElement()));
      }

      return res1;
    }

    Statement straighten(Statement st) throws ModelException {

      //      System.err.println("Straightening: " + st + " type of arc: " + st.predicate().getClass());

      if(st instanceof InvertedStatement)
	
	return (Statement)((InvertedStatement)st).getNonInverted();

      if(isInverted(st.predicate()))

	st = peer.getNodeFactory().createStatement(makeResource(st.object()), invert(st.predicate()), st.subject());

      //      System.err.println("   got: " + st + " type of arc: " + st.predicate().getClass());

      return st;
    }

    public void add(Statement st) throws ModelException {

      peer.add(straighten(st));
    }

    public void remove(Statement st) throws ModelException {

      peer.remove(straighten(st));
    }

    public boolean contains(Statement t) throws ModelException {

      return peer.contains(straighten(t));
    }

    public boolean isMutable() throws ModelException {

      return peer.isMutable();
    }

    public Model duplicate() throws ModelException {

      return new InvertedModel(peer.duplicate());
    }

    public Model create() throws ModelException {

      return new InvertedModel(peer.create());
    }

    public NodeFactory getNodeFactory() throws ModelException {

      return peer.getNodeFactory();
    }

    public Enumeration elements() throws ModelException {

      return new InvertedEnumeration(peer.elements());
    }
  }

  class InvertedEnumeration implements Enumeration {

    Enumeration en;
    Object nextInverted = null;

    public InvertedEnumeration(Enumeration en) {

      this.en = en;
    }

    public boolean hasMoreElements() {

      return nextInverted != null || en.hasMoreElements();
    }

    public Object nextElement() {

      if(nextInverted != null) {

	Statement res = new InvertedStatement((Statement)nextInverted);
	nextInverted = null;
	return res;

      } else {

	return (nextInverted = en.nextElement());
      }
    }
  }
}
