/******************************************************************************

* hull.c <s.hull>

* Creates the convex hull surrounding a sites list



* @Copyright Andrea Aime <aaime@libero.it>

* 23 Sept. 2001

* Last updated 19 Dec 2003, Markus Neteler to 5.7

*



* This file is part of GRASS GIS. It is free software. You can

* redistribute it and/or modify it under the terms of

* the GNU General Public License as published by the Free Software

* Foundation; either version 2 of the License, or (at your option)

* any later version.



* This program is distributed in the hope that it will be useful,

* but WITHOUT ANY WARRANTY; without even the implied warranty of

* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

* GNU General Public License for more details.

******************************************************************************/



#include <stdlib.h>

#include <stdio.h>

#include <assert.h>

#include "hull.h"



int rightTurn(struct Point *P, int i, int j, int k) {

    double a, b, c, d;

    a = P[i].x - P[j].x;

    b = P[i].y - P[j].y;

    c = P[k].x - P[j].x;

    d = P[k].y - P[j].y;

    return a*d - b*c < 0;

}





int cmpPoints(const void* v1, const void* v2) {

    struct Point *p1, *p2;

    p1 = (struct Point*) v1;

    p2 = (struct Point*) v2;

    if( p1->x > p2->x )

        return 1;

    else if( p1->x < p2->x )

        return -1;

    else

        return 0;

}





int convexHull(struct Point* P, int numPoints, int **hull) {

    int pointIdx, upPoints, loPoints;

    int *upHull, *loHull;



    /* sort points in ascending x order*/

    qsort(P, numPoints, sizeof(struct Point), cmpPoints);



    *hull = (int*) malloc(numPoints * sizeof(int));



    /* compute upper hull */

    upHull = *hull;

    upHull[0] = 0;

    upHull[1] = 1;

    upPoints = 1;

    for(pointIdx = 2; pointIdx < numPoints; pointIdx++) {

        upPoints++;

        upHull[upPoints] = pointIdx;

        while( upPoints > 1 &&

               !rightTurn(P, upHull[upPoints], upHull[upPoints-1],

                             upHull[upPoints-2])

             ) {

            upHull[upPoints-1] = upHull[upPoints];

            upPoints--;

        }

    }



    /*

    printf("upPoints: %d\n", upPoints);

    for(pointIdx = 0; pointIdx <= upPoints; pointIdx ++)

        printf("%d ", upHull[pointIdx]);

    printf("\n");

    */



    /* compute lower hull, overwrite last point of upper hull */

    loHull = &(upHull[upPoints]);

    loHull[0] = numPoints - 1;

    loHull[1] = numPoints - 2;

    loPoints = 1;

    for(pointIdx = numPoints - 3; pointIdx >= 0; pointIdx--) {

        loPoints++;

        loHull[loPoints] = pointIdx;

        while( loPoints > 1 &&

               !rightTurn(P, loHull[loPoints], loHull[loPoints-1],

                             loHull[loPoints-2])

             ) {

             loHull[loPoints-1] = loHull[loPoints];

             loPoints--;

        }

    }



    /*

    printf("loPoints: %d\n", loPoints);

    for(pointIdx = 0; pointIdx <= loPoints; pointIdx ++)

        printf("%d ", loHull[pointIdx]);

    printf("\n");

    */



    /* reclaim uneeded memory */

    *hull = (int*) realloc((char*)(*hull), (loPoints + upPoints) * sizeof(int));

    return loPoints + upPoints;

}







