/**
 *    '$RCSfile: Config.java,v $'
 *
 *     '$Author: ruland $'
 *       '$Date: 2005/12/08 14:28:18 $'
 *   '$Revision: 1.16 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.util;

import edu.ucsb.nceas.configxml.ConfigXML;
import edu.ucsb.nceas.configxml.exception.ElementNotFoundException;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.util.Hashtable;
import java.util.Vector;

import org.apache.xpath.XPathAPI;

import org.ecoinformatics.util.DebugPrinter;

import org.w3c.dom.NodeList;
import org.w3c.dom.Node;


/**
 *
 * configuration file routines
 *
 */
public class Config
{
  //private static final String configfilename = "kepler/config.xml";
  private static final String configfilename = "config.xml";

  private static final String settingsdir = ".kepler";

  // Define useful static strings.  At some point in time, these could be moved
  // into the config.xml file, environment variable, or something else
  // which could move this directory to another location.
  
  // KEPLER_USER_DIR is returned by Config.getUserDirPath().
  private static final String KEPLER_USER_DIR = System.getProperty("user.home")
                + File.separator + settingsdir + File.separator;

  // KEPLER_CACHE_DIR is returned by Config.getCacheDirPath().
  private static final String KEPLER_CACHE_DIR = KEPLER_USER_DIR + "cache" + File.separator;

  private static Config configure = null;
  private static DebugPrinter dbg = DebugPrinter.getInstance();
  private ConfigXML config;
  private boolean error = false;
  
  /**
   * Method for get object as a singleton class
   * @return Config
   */
  public static Config getInstance()
  {
    if (configure == null)
    {
      configure = new Config();
    }
    return configure;
  }

  /*
   * Constructor for ConfigDatasource.
   * @throws FileNotFoundException
   * @throws Exception
   */
  private Config()
  {
    try
    {
        InputStream is = this.getClass().getClassLoader().getResourceAsStream(configfilename);
        config = new ConfigXML(is);
        //config = new ConfigXML(configfilename);
    }
    catch(FileNotFoundException fnfe)
    {
      error = true;
      throw new RuntimeException("Config file not found at " +
          configfilename +
          " " + fnfe.getMessage());
    }
    catch(Exception e)
    {
      error = true;
      throw new RuntimeException("Unspecified error creating config object: " +
        e.getClass().getName() + "\n" + e.getMessage());
    }
  }

  /**
   * Method to get a string value from config.xml base on given element path
   * @param path String
   * @return String
   */
  public static String getValue(String path)
  {
    Config configInstance = Config.getInstance();
    return configInstance.getValueFromPath(path);
  }

  private String getValueFromPath(String path)
  {
    String value = null;
    try
    {
      if (!error && path != null)
      {
        Vector valueVector = config.getValuesForPath(path);
        value = (String) valueVector.elementAt(0);

      }

    }
    catch (Exception e)
    {
      throw new RuntimeException("Error to get value for path " + path +
                                 " in config.xml file " +e.getMessage());
    }
    dbg.print("The value for path " + path + " in config.xml is " + value, 2);
    return value;
  }//getValues

  /**
   * Method to get list (vector) values from config.xml base on given path
   * @param path String
   * @return Vector
   */
  public static Vector getList(String path)
  {
    Config configInstance = Config.getInstance();
    return configInstance.getListFromPath(path);
  }

  private Vector getListFromPath(String path)
  {
    Vector list = null;
    try
    {
      if (!error && path != null)
      {
        list = config.getValuesForPath(path);
      }
    }
    catch (Exception e)
    {
      throw new RuntimeException("Error to get list for path " + path +
                                 " in config.xml file " +e.getMessage());
    }

    return list;
  }//getList

  /**
   * Method to get map for some given pathes in config.xml
   * Here is the segment of config.xml:
   * <parenetElement>
   *   <keyElement>key1</keyElement>
   *   <valueElement>value1</valueElement>
   *  </parentElement>
   *  <parentElement>
   *   <keyElement>key2</keyElement>
   *   <valueElement>value2</valueElement>
   *  </parentElement>
   * The parentPath can be "//parentElement", keyPath should be "./keyElement",
   * and valuePath should be "./valuePath".
   * @param parentPath String
   * @param keyPath String
   * @param valuePath String
   * @throws ElementNotFoundException
   * @throws Exception
   * @return Hashtable
   */
  public static Hashtable getMap(String parentPath, String keyPath, String valuePath)
  {
    Config configInstance = Config.getInstance();
    return configInstance.getMapFromPath(parentPath, keyPath, valuePath);
  }

  private Hashtable getMapFromPath(String parentPath, String keyPath, String valuePath)
  {
    Hashtable map = new Hashtable();
    try
    {
      if (!error && parentPath != null && keyPath != null && valuePath != null)
      {
        NodeList nl = config.getPathContent(parentPath);
        for (int i = 0; i < nl.getLength(); i++)
        {
          Node n = nl.item(i);
          Node keyNode = XPathAPI.selectSingleNode(n, keyPath);
          Node valueNode = XPathAPI.selectSingleNode(n, valuePath);
          if (keyNode == null || valueNode == null)
          {
            throw new RuntimeException("Expecting a " + keyPath + " node or " +
                                       valuePath + " node " + "as child of  " +
                                       parentPath +
                                       "could not found in config.xml.");
          }

          String keyNodeValue = keyNode.getFirstChild().getNodeValue();
          String valueNodeValue = valueNode.getFirstChild().getNodeValue();
          dbg.print("The value of " + keyPath + " " + keyNodeValue + " and " +
                    " the value of " + valuePath + " " + valueNodeValue +
                    " are put into mapping", 2);

          if (keyNodeValue == null || valueNodeValue == null)
          {
            throw new RuntimeException("The value of " + keyPath + " or " +
                                       valuePath + " is null ");
          }
          map.put(keyNodeValue, valueNodeValue);
        } //for
      } //if
    }//try
    catch(Exception e)
    {
      throw new RuntimeException("Error to get map for path " + parentPath + " "
                                 + keyPath + " " + valuePath +
                                 " in config.xml file " +e.getMessage());
    }//catch
    return map;
  }//getMap
  
  /**
   * Get a node list base on the given xpath in config file
   * @param path String  the xpath
   * @return NodeList
   */
  public static NodeList getNodeListFromPath(String path)
  {
    Config configInstance = Config.getInstance();
    return configInstance.getNodeListFromXPath(path);
  }
  
  /*
   * This method will get a node list from config file
   */
  private NodeList getNodeListFromXPath(String xpath)
  {
    NodeList list = null;
    try
    {
      list = config.getPathContent(xpath);
    }
    catch (Exception e)
    {
      throw new RuntimeException("Error to get nodelist for path " + xpath +
                                 e.getMessage());
    }
    return list;
  }

  /**
   * Method to get error state
   */
  public boolean getErrorState()
  {
    return error;
  }

  /**
   * Method getDebugLevel.
   * @return String
   */
  public int getInputStreamBufferSize()
  {
    if( !error ) {
      try {
        return new Integer(config.get("InputStreamBufferSize", 0).trim()).intValue();
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException e)
      {
        System.out.println("You need to add a 'InputStreamBufferSize' element to your " +
          "config file.  InputStreamBufferSize is defaulting to 1024.");
      }
    }
    return 1024;
  }

  /** 
   * Returns the path of the Kepler hidden working directory terminated with File.seperator.
   * This directory is current coded to be ${system.home}/.kepler.
   * 
   * @return the path to the .kepler directory.
   */
  public static String getUserDirPath()
  {
	  return KEPLER_USER_DIR;  
  }
  
  /**
   * Returns the path to the Cache working directory terminated with File.seperator.
   * This directory is currently in getUserDirPath() + "cache".
   * 
   * @return the path to the cache directory
   */
  public static String getCacheDirPath()
  {
	  return KEPLER_CACHE_DIR;
  }
  
  /**
   * Method getCatalogPath.
   * @return String
   */
  public String getCatalogPath() throws Exception
  {
    if( !error ) {
      try {
        return config.get("CatalogPath", 0);
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException e)
      {
        throw new Exception("Catalog path element not found");
      }
    }
    return(null);
  }

  /**
   * Method getTempDir.
   * @return String
   */
  public String getTempDir()
  {
    if( !error ) {
      try {
        return config.get("TempDir", 0);
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException e)
      {
        //        System.out.println("TempDir element not found");
      }
    }
    return(null);
  }

  /**
   * Method getSAXParser.
   * @return String
   */
  public String getSAXParser()
  {
   if( !error ) {
      try {
        return config.get("SAXParser", 0);
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException e)
      {
        System.out.println("SAXParser element not found in config file");
      }
    }
    return(null);
  }

  /**
   * Method getHtmlStartPage.
   * @return Object
   */
  public String getHtmlStartPage() throws Exception
  {
    if( !error ) {
      try {
        return config.get("HtmlStartPage", 0);
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException e)
      {
        throw new Exception("HtmlStartPage element not found in config file");
      }
    }
    return("");
  }

  /**
   * Method getLangNodeList.
   * @return NodeList
   * @throws ExecutionException
   */
  public NodeList getLangNodeList()
    throws Exception
  {
    if( !error )
    {
      try
      {
        return config.getPathContent("/config/language");
      }
      catch(Exception e)
      {
        throw new Exception("Could not get language attribute from " +
                                   "the config file.");
      }
    }
    return null;
  }

  /**
   * Method getDataTypes.
   * @return NodeList
   * @throws TransformerException
   */
  public NodeList getDataTypes() throws Exception
  {
    if( !error ) {
      try
      {
        return config.getPathContent("//dataType");
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException enfe)
      {
        throw new Exception("Cannot find any dataType elements " +
          "in the config.xml file: " + enfe.getMessage());
      }

    }
    return null;
  }

  public NodeList getDBdataTypeMappings() throws Exception
  {
    if( !error ) {
      try
      {
        return config.getPathContent("//DBdataTypeMapping/mapping");
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException enfe)
      {
        throw new Exception("Cannot find any data type mapping elements " +
          "in the config.xml file: " + enfe.getMessage());
      }

    }
    return null;

  }

  /**
   * A method to get delimiter mapping nodelist in config file
   * @throws Exception
   * @return NodeList
   */
  public NodeList getDelimiterMappings() throws Exception
  {
    if( !error ) {
      try
      {
        return config.getPathContent("//delimiterMapping/mapping");
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException enfe)
      {
        throw new Exception("Cannot find any delimiter mapping elements " +
          "in the config.xml file: " + enfe.getMessage());
      }

    }
    return null;

  }

  /**
   * returns the lineEnding element
   */
  public String getLineEnding()
  {
    if( !error )
    {
      try
      {
        return config.get("lineEnding", 0);
      }
      catch(edu.ucsb.nceas.configxml.exception.ElementNotFoundException e)
      {
        throw new RuntimeException("lineEnding element not found in " +
           "config.xml file.");
      }
    }
    return "\n";
  }

  /**
   * Method getLocalCatalogPath.
   * @return String
   * @throws Exception
   */
  public String getLocalCatalogPath() throws Exception
  {
    try
    {
      return config.get("local_catalog_path", 0);
    }
    catch(Exception e)
    {
      throw new Exception("Error getting local_catalog_path: " +
                        e.getMessage());
    }
  }

  /**
   * Method save. - pass through to ConfigXML member
   * @throws Exception
   */
  public void save() throws Exception
  {
    config.save();
  }

  /**
   * Method set. - pass through to ConfigXML member
   * @param arg1
   * @param arg2
   * @param arg3
   * @throws ElementNotFoundException
   */
  public void set(String arg1, int arg2, String arg3) throws ElementNotFoundException
  {
    config.set(arg1, arg2, arg3);
  }

}

