/**
 *    '$RCSfile: DelimitedReader.java,v $'
 *
 *     '$Author: tao $'
 *       '$Date: 2005/03/19 01:02:29 $'
 *   '$Revision: 1.13 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.util;

import java.io.*;
import java.util.*;

/**
 * tokenizes a delimited file.  This reader assumes that one record is on one
 * line which ends with the line
 */
public class DelimitedReader
{
  private String data;
  private InputStreamReader dataReader;
  private Vector[] lines;
  private Vector linesVector;
  private int numHeaderLines;
  private int numRecords;
  private boolean stripHeader = false;
  private int currentRecord = 0;
  private int numCols;
  private String delimiter;
  private String lineEnding;
  private static org.ecoinformatics.util.DebugPrinter dbg = org.ecoinformatics.util.DebugPrinter.getInstance();;

  /**
   * constructor. reads the csv stream.
   * @param delimString the delimited stream to read
   * @param numCols the number of columns in the stream
   * @param delimiter the delimiter to tokenize on
   * @param numHeaderLines the number of lines to skip at the top of the file
   * @param lineEnding the line ending char(s)...either "\n" (unix),
   * "\r\n" (windoze) or "\r" (mac)
   */
  public DelimitedReader(String data, int numCols, String delimiter,
    int numHeaderLines, String lineEnding, int numRecords)
  {
    this.numHeaderLines = numHeaderLines;
    this.data = data;
    this.numCols = numCols;
    this.numRecords = numRecords;    
    dbg.print("Delimiter is: " + delimiter, 2);
    this.delimiter = unescapeDelimiter(delimiter);
    dbg.print("LineEnding is: " + lineEnding, 2);
    this.lineEnding = unescapeDelimiter(lineEnding);
    
    //lines = new Vector[numRecords + numHeaderLines + 1];
    linesVector = new Vector();

    int begin = 0;
    int end = 0;
//    int i = 0;
    while(end < data.length())
    { //add each line of the string as an element in a vector
      end = data.indexOf(this.lineEnding, begin); //DFH 'this.' added
      if(end == -1)
      {
        end = data.length();
      }
      String line = data.substring(begin, end);
      if(!line.trim().equals(""))
      {
        //take off the line ending
        // MBJ: I commented out the next line as it was improperly truncating lines
        // I'm not sure why it was there in the first place, as the previous substring
        // removed the delimiter
        //line = line.substring(0, line.length() - lineEnding.length());

        //split the line based on the delimiter
        Vector v = splitDelimitedRowStringIntoVector(line);
        /*String[] s = line.split(delimiter.trim(), numCols);
        Vector v = new Vector();
        for(int j=0; j<s.length; j++)
        {
          v.addElement(s[j]);
        }

        if(v.size()  < numCols)
        {
          int vsize = v.size();
          for(int j=0; j<numCols - vsize; j++)
          { //add any elements that aren't there so that all the records have the
            //same number of cols
            v.addElement("");
          }
        }*/
        //lines[i] = v;
        linesVector.add(v);
//        i++;
      }
      //go to the next line
      begin = end + this.lineEnding.length();   //DFH  'this.' added
    }
    
    int records = linesVector.size();
    if (records != this.numRecords) {
        this.numRecords = records;
        dbg.print("Metadata disagrees with actual data. Changing number of records to: " + records, 1);
    }
    lines = new Vector[records];
    for (int k=0; k < records; k++) {
        lines[k] = (Vector)linesVector.get(k);
    }
/*
    for(int j=0; j<lines.length; j++)
    {
      if(lines[j] == null)
      {
        lines[j] = new Vector();
      }
    }
*/
    
  }
  
  /**
   * This constructor will read delimitered data from stream rather a string
   * @param dataStream InputStream  The input stream
   * @param numCols int  the number of columns
   * @param delimiter String  delimiter the delimiter to tokenize on
   * @param numHeaderLines int numHeaderLines the number of lines to skip at the top of the file
   * @param lineEnding String  lineEnding the line ending char(s)...either "\n" (unix),"\r\n" (windoze) or "\r" (mac)
   * @param numRecords int number of rows in the input stream
   */
  public DelimitedReader(InputStream dataStream, int numCols, String delimiter,
    int numHeaderLines, String lineEnding, int numRecords, boolean stripHeader)
  {
    this.dataReader = new InputStreamReader(dataStream);
    this.numHeaderLines = numHeaderLines;
    this.numCols = numCols;
    this.numRecords = numRecords;    
    dbg.print("Delimiter is: " + delimiter, 2);
    this.delimiter = unescapeDelimiter(delimiter);
    dbg.print("LineEnding is: " + lineEnding, 2);
    this.lineEnding = unescapeDelimiter(lineEnding);
    this.stripHeader = stripHeader;

  }
 
 /**
  * Method to set up data stream as source
  * @param dataStream InputStream
  */
  public void setInputStream(InputStream dataStream)
  {
    this.dataReader = new InputStreamReader(dataStream);
  }
  
  /**
   * This method is from data source as a input stream
   * This method will read one row from and return a data vector which element is String
   * and the value is field data. After reach the end of stream, empty vector 
   * will be returned. So this method can be iterated by a while loop until
   * a empty vector hited. During the iteration, every data in the stream will
   * be pulled out.
   * @return Vector
   */
  public Vector getRowDataVectorFromStream() 
  {
    Vector oneRowDataVector = new Vector();
    StringBuffer rowData = new StringBuffer();
    String rowDataString = null;
    int singleCharactor = -2;
    int headLineNumberCount = 0;
    if (dataReader != null)
    {
      try
      {
        while (singleCharactor != -1)
        {
          singleCharactor = dataReader.read();
          char charactor = (char)singleCharactor;
          rowData.append(charactor);
          // find string - line ending in the row data
          if (rowData.indexOf(lineEnding) != -1)
          {
            // strip the header lines
            if (stripHeader && numHeaderLines >0 && headLineNumberCount < numHeaderLines)
            {
               // reset string buffer(descard the header line)
               rowData = null;
               rowData = new StringBuffer();
               headLineNumberCount++;
            }
            else
            {
              rowDataString = rowData.toString();
              dbg.print("The row data is " + rowDataString, 2);
              break;
            }
          }
        }
      }
      catch (Exception e)
      {
        dbg.print("Couldn't read data from input stream", 2);
        oneRowDataVector = new Vector();
      }
    }
    if (rowDataString != null)
    {
      oneRowDataVector = splitDelimitedRowStringIntoVector(rowDataString);
    }
    return oneRowDataVector;
  }
  
  /*
   * This method will read a delimitered string and put a delimitered part into
   * an element in a vector. If the vector size is less than the column number
   * empty string will be added.
   */
  private Vector splitDelimitedRowStringIntoVector(String data)
  {
    Vector result = new Vector();
    if (data == null)
    {
      return result;
    }
    String[] s = data.split(delimiter.trim(), numCols);
    if( s != null)
    {
        for (int j = 0; j < s.length; j++) 
        {
          if (s[j] != null)
          {
            result.addElement(s[j].trim());
          }
          else
          {
              result.addElement("");
          }
        }
        //add any elements that aren't there so that all the records have the
        //same number of cols
        if (result.size() < numCols) 
        {
          int vsize = result.size();
          for (int j = 0; j < numCols - vsize; j++) 
          {
            result.addElement("");
          }
        }
    }
    return result;
  }
  
  
  
  /**
   * returns the data as an array of vectors.  each vector will have the same
   * number of elements as there are columns in the data.
   * @param stripHeaderLines true if the header lines should not be included
   * in the returned data, false otherwise
   */
  public Vector[] getTokenizedData(boolean stripHeaderLines)
  {
    if(stripHeaderLines)
    {
      Vector[] strip = null;
      if (numRecords > numHeaderLines)
      {
        strip = new Vector[numRecords-numHeaderLines];
        for(int i=numHeaderLines; i<lines.length; i++)
        {
          strip[i-numHeaderLines] = lines[i];
        }
      }
      return strip;
    }
    else
    {
      return lines;
    }
  }

  /**
   * returns a string representation of the data
   */
  public String toString()
  {
    StringBuffer sb = new StringBuffer();
    for(int i=0; i<lines.length; i++)
    {
      dbg.print("line[" + (i + 1) + "]: " + lines[i].toString(), 2);
      for(int j=0; j<lines[i].size(); j++)
      {
        sb.append((String)lines[i].elementAt(j));
        if(j != lines[i].size() - 1)
        {
          sb.append(" || ");
        }
      }
      sb.append(lineEnding);
    }
    return sb.toString();
  }
  
  /**
   * Convert a string escaped representation of a delimiter character into
   * an the actual String for that delimiter.  This is used for translating
   * escaped versions of tab, newline, and carriage return characters to
   * their real character values.
   * 
   * @param delimiter the String representing the delimiter
   * @return the actual String for the delimiter
   */
  public static String unescapeDelimiter(String delimiter)
  {
      String newDelimiter = delimiter;
      
      if (delimiter == null) {
          dbg.print("Delimiter is null and we set up to \n.", 2);
          newDelimiter = "\n";   
      }else if (delimiter.equals("\\t")) {
          dbg.print("Tab interpreted incorrectly as string.", 2);
          newDelimiter = "\t";
      } else if (delimiter.equals("\\n")) {
          dbg.print("Newline interpreted incorrectly as string.", 2);
          newDelimiter = "\n";
      } else if (delimiter.equals("\\r")) {
          dbg.print("CR interpreted incorrectly as string.", 2);
          newDelimiter = "\r";
      } else if (delimiter.equals("\\r\\n")) {
          dbg.print("CRNL interpreted incorrectly as string.", 2);
          newDelimiter = "\r\n";
      } else if (delimiter.startsWith("#")){
          dbg.print("XML entity charactor.", 2);
          String digits = delimiter.substring(1, delimiter.length());
          int radix = 10;
          if (digits.startsWith("x"))
          {
              dbg.print("Radix is "+ 16, 2);
              radix = 16;
              digits = digits.substring(1, digits.length());
          }
          dbg.print("Int value of  delimiter is "+digits, 2);
          Integer integer = Integer.valueOf(digits, radix);
          int inter = integer.intValue();
          dbg.print("The decimal value of char is "+ inter, 2);
          char charactor =(char)inter;
          newDelimiter = Character.toString(charactor);
          dbg.print("The new delimter is "+newDelimiter, 2);
      }
      

      return newDelimiter;
  }
}
