/**
 *    '$RCSfile: DataCacheFileObject.java,v $'
 *
 *     '$Author: berkley $'
 *       '$Date: 2005/12/02 19:15:51 $'
 *   '$Revision: 1.6 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2003 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
package org.kepler.objectmanager.cache;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;

import org.ecoinformatics.seek.querybuilder.DBUIUtils;

import org.kepler.objectmanager.cache.CacheUtil;

import org.w3c.dom.Node;

/**
 * This class is not as generic as the name may indicate. It is designed to get the metadata to determine
 * where the data is stored and then does a Ecogrid "get" to the data.
 */
public class DataCacheFileObject extends DataCacheObject
{
    // File Type
    public static final int UNKNOWN            = 0; // file stored internally
    public static final int INTERNAL           = 1; // file stored internally
    public static final int EXTERNAL           = 2; // file is external to cache
    public static final int COPY_FILE_TO_CACHE = 3; // file is moved, then Type is set to INTERNAL

    // Protected
    protected int       mFileLoc   = UNKNOWN;
    protected File      mFile       = null;

    /**
     * A DataCacheFileObject object that can point to an itnernal file or external file object
     *
     */
    public DataCacheFileObject()
    {
        super();
    }

    /**
     * @return Returns whether the file is internal or external, if nothing has happened then it return UNKNOWN
     */
    public int getFileLocation()
    {
        return mFileLoc;
    }

    /**
     * @return Returns the string describing the files type
     */
    public String getType()
    {
        return mResourceName;
    }

    /**
     * Set the status of the file item
     * @param aFileLoc the location type
     */
    public void setFileLocation(int aFileLoc)
    {
        mFileLoc = aFileLoc;
    }

    /**
     * Return the File object for this cache item, the File object points to
     * "physical" data file eithe rin the cache of externally
     * @return java.io.File or null
     */
    public File getFile()
    {
        if (mFile == null)
        {
            mFile = new File(mLocalFileName);
        }
        return mFile;
    }

    /**
     * Return the File object as a FileInputStream
     * @return java.io.FileInputStream
     */
    public FileInputStream getInputStream()
      throws IOException
    {
        if (mFile == null)
        {
            mFile = new File(mLocalFileName);
        }
        return new FileInputStream(mFile);
    }

    /**
     * This will initialize the current object with Name and Resource name,
     * create and save the cache item with the data that is passed in from ther byte array
     * @param aFileName the file name, this is only a textual identifier and has nothing to do
     *        with the actual file name (although the caller may put the actual external file name in theis arg)
     * @param aResName The resource name is not used for anything except making this item uniquely named
     * @param aData the data to be persisted
     */
    public void initializeWithBytes(String aName, String aType, byte[] aData)
    {
        mName         = aName;
        mResourceName = aType;
        mData         = new byte[aData.length];
        mSize         = aData.length;
        for (int i=0;i<aData.length;i++)
        {
            mData[i] = aData[i];
        }
        mHasChanged = true;
        this.saveData();
        mFile = new File(mLocalFileName);

    }

    /**
     * This copies the external file into the cache and saves it
     * @param aPhysicalFileName the physical file name to be copied
     */
    private void copyFileToCache(String aPhysicalFileName)
    {
        File currentFile = new File(aPhysicalFileName);
        if (currentFile != null && currentFile.exists())
        {
            try {
                mLocalFileName = createFileName();
                mFile = new File(mLocalFileName);
                if (mFile != null) {
                    FileOutputStream fos = new FileOutputStream(mFile);
                    if (fos != null)
                    {
                       BufferedOutputStream bos = new BufferedOutputStream(fos);
                       if (bos != null)
                       {
                           FileInputStream fis  = new FileInputStream(currentFile);
                           if (fis != null)
                           {
                               BufferedInputStream bis = new BufferedInputStream(fis);
                               if (bis != null)
                               {
                                   long filesize = currentFile.length();
                                   int size = 20500;
                                   byte[] data = new byte[size]; // 20K
                                   while (filesize > 0)
                                   {
                                       int retSize = bis.read(data, 0, size);
                                       bos.write(data, 0, retSize);
                                       filesize -= retSize;
                                   }
                                   mFileLoc = INTERNAL;
                                   mStatus = CACHE_OK;
                                   mSize = (int)filesize;       // XXX is this a bad thing to do?
                                   bis.close();
                               }
                               fis.close();
                           }
                           bos.close();
                       }
                       fos.close();
                    }
                }
            } catch (Exception e) {
               System.err.println(e);
               mFile          = null;
               mLocalFileName = null;
               mStatus        = CACHE_ERROR;
               mFileLoc      = UNKNOWN;
            }
        }

    }

    /**
     * This is one of the hardest working methods in the class. It determines what to do
     * the aFileName argument. If the file type is:
     * INTENRAL then it creates a new "empty" file, and the aFileName is just text
     * EXTERNAL then it uses aFileName as the mLocalFileName data member in order to point at the extneral
     *          file. NOTE: An extneral file is NEVER deleted by the DCM
     * COPY_FILE_TO_CACHE then it copies the file from the extneral location identifed by the aFileName argument
     *          to an intneral cache file. Afterward it sets the mFileLoc to INTERNAL
     *
     * @param aPhysicalFileName the "physical" name of a file (this name will not be stored if aFileLocation is COPY_FILE_TO_CACHE)
     * @param aLogicalName the "logical" name of a file unrelated to the physical name
     * @param aType a description of the file's type (optional and can be null)
     * @param aFileLocation the file location type (see DataCacheFileObject for values)
     */
    public void initializeWithFileName(String aPhysicalName, String aLogicalName, String aType, int aFileLocation)
    {
        if (mFileLoc == INTERNAL)
        {
            if (mFile != null)
            {
                mFile.delete();
            }
        }

        mFileLoc      = aFileLocation;
        mFile          = null;
        mLocalFileName = null;
        mData          = null;
        mName          = aLogicalName;
        mResourceName  = aType;

        switch (mFileLoc)
        {
            case INTERNAL :
                mLocalFileName = createFileName();
                mFile          = new File(mLocalFileName);
                mStatus        = CACHE_OK;
                mSize          = (int)mFile.length();
                break;

            case EXTERNAL :
                mLocalFileName = aPhysicalName;
                mFile          = new File(mLocalFileName);
                mStatus        = CACHE_OK;
                mSize          = (int)mFile.length();
                break;

            case COPY_FILE_TO_CACHE :
                copyFileToCache(aPhysicalName);
                break;
        }
    }
    
    /**
     * create a new DataCacheFileObject with data from an InputStream.  The
     * physical file name in the cache will be the same as the logical name.
     * an IOException will be thrown if anything goes wrong with writing the
     * inputstream to disk (including filename problems, so be careful what
     * you put in your filename (i.e. no ':/#%', etc.))  All items will be added
     * as INTERNAL
     *
     * @param aFileInputStream the stream to read the data from
     * @param aLogicalName the "logical" name of a file unrelated to the physical name
     * @param aPhysicalName the physical file name
     */
    public void initializeWithInputStream(String aLogicalName,
      InputStream aFileInputStream) throws IOException
    {
      initializeWithInputStream(aLogicalName, null, aFileInputStream);
    }

    /**
     * create a new DataCacheFileObject with data from an InputStream.  The
     * physical file name in the cache will be the same as the logical name.
     * an IOException will be thrown if anything goes wrong with writing the
     * inputstream to disk (including filename problems, so be careful what
     * you put in your filename (i.e. no ':/#%', etc.))  All items will be added
     * as INTERNAL
     *
     * @param aFileInputStream the stream to read the data from
     * @param aLogicalName the "logical" name of a file unrelated to the physical name
     * @param aPhysicalName the physical file name
     */
    public void initializeWithInputStream(String aLogicalName, String aType,
      InputStream aFileInputStream)
      throws IOException
    {
        mFile          = null;
        mFileLoc       = INTERNAL;
        mLocalFileName = null;
        mData          = null;
        mName          = aLogicalName;
        mResourceName  = aType;

        mLocalFileName = createFileName();
        mFile          = new File(mLocalFileName);
        FileOutputStream fos = new FileOutputStream(mFile);
        CacheUtil.writeInputStreamToOutputStream(aFileInputStream, fos);
        mStatus        = CACHE_OK;
        mSize          = (int)mFile.length();
    }

    //----------------------------------------------------------------
    //-- Overrides
    //----------------------------------------------------------------

    /**
     *
     */
    protected String getSerialClassName()
    {
      return DataCacheFileObject.class.getName();
    }

    /* (non-Javadoc)
     * @see org.ecoinformatics.seek.datasource.DataCacheObject#doWork()
     */
    public void doWork()
    {
        if (mFileLoc != UNKNOWN)
        {
            if (this.mLocalFileName != null && mLocalFileName.length() > 0)
            {
                mFile = new File(mLocalFileName);
                if (mFile.exists())
                {
                    mStatus = CACHE_OK;
                }
            }
        }
    }

    /**
     * Return and extra DataCacheObject attributes
     * @return
     */
    protected String getAttrsForXML()
    {
        StringBuffer strBuf = new StringBuffer();
        strBuf.append("fileType=\"" + mFileLoc +"\"\n");
        return strBuf.toString();
    }

    /**
     * Read in any of the extra attributes
     * @param aNode the DOM node to use to search for the attributes(s)
     */
    public boolean buildFromDOM(Node aNode)
    {
        if (aNode != null)
        {
            mFileLoc = getIntFromStr(DBUIUtils.findAttrValue(aNode, "fileType"));
            if (mFileLoc == Integer.MAX_VALUE)
            {
                mFileLoc = UNKNOWN;
                return false;
            }
        }
        boolean ok = super.buildFromDOM(aNode);
        if (ok)
        {
            mFile = new File(mLocalFileName);
            mSize = (int)mFile.length();       // XXX is this a bad thing to do?
            mStatus = CACHE_OK;
        }
        return ok;
    }

    /**
     * There really isn't any notion of saving data that is in memory, because this object is really
     * a reference to a physical location, so there is nothing ever to save.
     */
    public boolean saveData()
    {
        return true;
    }

    /**
     * An overridable method for enabling the item to determine if
     * it can be persisted into the cache
     * @return true if it can be cached, false if not
     */
    protected boolean isOKToPersist()
    {
        return (mLocalFileName != null &&
                mName.length() > 0 &&
                mFileLoc != UNKNOWN &&
                mStatus == CACHE_OK);
    }

    /**
     * Clear the data from this cache item and remove the file if it is internal
     *
     */
    public void clear()
    {
        if (mLocalFileName != null &&
            mLocalFileName.length() > 0 &&
            mFileLoc == INTERNAL)
        {
            File file = new File(mLocalFileName);
            if (file != null && file.exists())
            {
                file.delete();
            }
        }
        mData = null;
        removeAllListeners();
    }

}
