/**
 *   '$RCSfile: LSIDTree.java,v $'
 *
 *   '$Author: berkley $'
 *   '$Date: 2005/12/14 20:40:37 $'
 *   '$Revision: 1.11 $'
 *
 *  For Details: http://kepler-project.org
 *  Copyright (c) 2003-2004 The Regents of the University of California.
 *  All rights reserved.
 *  Permission is hereby granted, without written agreement and without
 *  license or royalty fees, to use, copy, modify, and distribute this
 *  software and its documentation for any purpose, provided that the above
 *  copyright notice and the following two paragraphs appear in all copies
 *  of this software.
 *  IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 *  FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 *  ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 *  THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 *  THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 *  PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 *  CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 *  ENHANCEMENTS, OR MODIFICATIONS.
 *  PT_COPYRIGHT_VERSION_2
 *  COPYRIGHTENDKEY
 */
package org.kepler.objectmanager.lsid;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.Writer;
import java.io.BufferedWriter;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Set;
import java.util.Map;
import java.util.Vector;

import org.ecoinformatics.util.Config;

import com.ibm.lsid.LSIDException;

/**
 * this  class stores and serializes a tree of lsids and allows
 * easy searching of the tree.  The tree is structured like:
 * authority
 *    |_ namespace
 *           |_ object
 *                 |_revision
 */
public class LSIDTree
{
  private Hashtable authTree;
  private static LSIDTree lsidTree = null;
  private File saveFile = new File( Config.getUserDirPath() + "LSIDTree");

  public LSIDTree()
  {
    authTree = new Hashtable();
  }

  /**
   * construct a new LSIDTree. Read the serialized version from disk if
   * the boolean is true.
   */
  public LSIDTree(boolean readFromDisk)
    throws IOException
  {
    authTree = new Hashtable();
    if(readFromDisk)
    {
      readTree();
    }
  }

  /**
   * return the singleton instance of this class
   */
  public static synchronized LSIDTree getInstance()
  {
    if(lsidTree == null)
    {
      try
      {
        lsidTree = new LSIDTree(true);
      }
      catch(IOException ioe)
      {
        lsidTree = new LSIDTree();
      }
    }
    return lsidTree;
  }

  /**
   * add the specified lsid to the tree.  If it already exists in the tree
   * throw an Exception.  If addtoTree is false, the lsid will be checked for
   * uniqueness but will not be added to the tree for future reference.
   */
  public void addLSID(KeplerLSID lsid) throws LSIDTreeException
  {
    int lastObj = -1;
    int lastRev = -1;

    Hashtable nsTree = (Hashtable)authTree.get(lsid.getAuthority());
    if(nsTree == null)
    {
      Hashtable revHash = new Hashtable();
      revHash.put(lsid.getRevision(), new Integer(-1));
      Hashtable objHash = new Hashtable();
      objHash.put(lsid.getObject(), revHash);
      nsTree = new Hashtable();
      nsTree.put(lsid.getNamespace(), objHash);
      authTree.put(lsid.getAuthority(), nsTree);
    }
    else
    {
      Hashtable objTree = (Hashtable)nsTree.get(lsid.getNamespace());
      if(objTree == null)
      {
        objTree = new Hashtable();
        Hashtable revHash = new Hashtable();
        revHash.put(lsid.getRevision(), new Integer(-1));
        objTree = new Hashtable();
        objTree.put(lsid.getObject(), revHash);
        nsTree.put(lsid.getNamespace(), objTree);
      }
      else
      {
        Enumeration objEnum = objTree.keys();
        while(objEnum.hasMoreElements())
        {
          int i = new Integer((String)objEnum.nextElement()).intValue();
          if(i > lastObj)
          { //find the highest number object
            lastObj = i;
          }
        }

        Hashtable revTree = (Hashtable)objTree.get(lsid.getObject());
        if(revTree == null)
        {
          revTree = new Hashtable();
          revTree.put(lsid.getRevision(), new Integer(-1));
          objTree.put(lsid.getObject(), revTree);
        }
        else
        {
          Enumeration revEnum = revTree.keys();
          Vector revs = new Vector();
          while(revEnum.hasMoreElements())
          {
            int rev = new Integer((String)revEnum.nextElement()).intValue();
            if(rev > lastRev)
            {
              lastRev = rev;
            }
            revs.addElement(new Integer(rev));
          }

          if(!revs.contains(new Integer(lsid.getRevision())))
          {
            revTree.put(lsid.getRevision(), new Integer(-1));
          }
          else
          { //if the revision already exists, throw an exception
            throw new LSIDTreeException(lsid, lastObj, lastRev);
          }
        }
      }
    }
  }
  
  /**
   * return, but do not register the next revision of the given lsid
   * @param lsid the lsid to find the next rev for.
   */
  public KeplerLSID findNextRevision(KeplerLSID lsid)
  {
    int lastRev = -1;

    Hashtable nsTree = (Hashtable)authTree.get(lsid.getAuthority());
    if(nsTree == null)
    {
      return lsid;
    }
    else
    {
      Hashtable objTree = (Hashtable)nsTree.get(lsid.getNamespace());
      if(objTree == null)
      {
        return lsid;
      }
      else
      {
        Hashtable revTree = (Hashtable)objTree.get(lsid.getObject());
        if(revTree == null)
        {
          return lsid;
        }
        else
        {
          Enumeration revEnum = revTree.keys();
          Vector revs = new Vector();
          while(revEnum.hasMoreElements())
          {
            int rev = new Integer((String)revEnum.nextElement()).intValue();
            if(rev > lastRev)
            {
              lastRev = rev;
            }
            revs.addElement(new Integer(rev));
          }

          try
          {
            return new KeplerLSID(lsid.getAuthority(), lsid.getNamespace(),
              new Integer(lsid.getObject()).intValue(), lastRev + 1);
          }
          catch(LSIDException lside)
          {
            throw new RuntimeException("Error finding last revision: " + 
              lside.getMessage());
          }
        }
      }
    }
  }
  
  /**
   * returns the lsid with the next possible revision number.  if you pass
   * in urn:lsid:localhost:actor:1:1, this should return
   * urn:lsid:localhost:actor:1:2 if 2 is the next number available to use.
   * if there is already a revision 2, it will return the next available
   * revision (i.e. 3).  if this lsid does not exist in the database,
   * the lsid will be registered and the same lsid will be returned to you.
   * Note that the new lsid WILL be registered in the database prohibiting
   * the same lsid from being registered again.
   * @param lsid the lsid to inc the revision
   
   */
  public KeplerLSID getNextRevision(KeplerLSID lsid)
  {
    return getNextRevision(lsid, true);
  }

  /**
   * returns the lsid with the next possible revision number.  if you pass
   * in urn:lsid:localhost:actor:1:1, this should return
   * urn:lsid:localhost:actor:1:2 if 2 is the next number available to use.
   * if there is already a revision 2, it will return the next available
   * revision (i.e. 3).  if this lsid does not exist in the database,
   * the lsid will be registered and the same lsid will be returned to you.
   * if register is false, the new lsid will not be registered in the database
   * @param lsid the lsid to inc the revision
   * @param register set to true if you want the new lsid registered, false
   * otherwise
   */
  public KeplerLSID getNextRevision(KeplerLSID lsid, boolean register)
  {
    try
    {
      if(register)
      {
        addLSID(lsid);
        return lsid;
      }
      else
        return findNextRevision(lsid);
      
    }
    catch(LSIDTreeException e)
    {
      try
      {
        KeplerLSID newLsid = new KeplerLSID("urn:lsid:" + lsid.getAuthority() + ":" +
          lsid.getNamespace() + ":" + lsid.getObject() + ":" +
          (e.lastRevision + 1));
        if(register)
          addLSID(newLsid);
        
        return newLsid;
      }
      catch(LSIDException lside)
      { //basically ignore this.  it shouldn't happen
        lside.printStackTrace();
        return null;
      }
    }
  }

  /**
   * return the KeplerLSID with the next object number or
   * return the passed lsid itself if it has not been registered
   * and is already has the next object number in the series
   * if register is true, register the new lsid.
   * @param lsid
   */
  public KeplerLSID getNextObject(KeplerLSID lsid, boolean register)
    throws LSIDException
  {
    int nextObjNum = getNextObject(lsid.getAuthority(), lsid.getNamespace());

    if(nextObjNum > (new Integer(lsid.getObject())).intValue())
    {
      KeplerLSID newLsid = new KeplerLSID(lsid.getAuthority(), lsid.getNamespace(),
        nextObjNum, 1);
      if(register)
      {
        addLSID(newLsid);
      }
      return newLsid;
    }

    addLSID(lsid);
    return lsid;
  }
  
  /**
   * get the next object and register the lsid
   */
  public KeplerLSID getNextObject(KeplerLSID lsid)
    throws LSIDException
  {
    return getNextObject(lsid, true);
  }

  /**
   * return a string rep of the tree.
   */
  public String toString()
  {
    return authTree.toString();
  }

  /**
   * save the tree to disk so it can be read back using readTree()
   */
  public synchronized void saveTree()
  throws IOException
  {
	  final String[] vectorTemplate = new String[0];
	  
	  Writer fw = new BufferedWriter( new FileWriter(saveFile) );
	  
	  Set authTreeEntries = authTree.entrySet();
	  Iterator entries = authTreeEntries.iterator();
	  while(entries.hasNext())
	  {
		  Map.Entry entry = (Map.Entry) entries.next();
		  
		  String authKey = (String)entry.getKey();
		  Hashtable nsTree = (Hashtable)entry.getValue();
		  
		  Set nsTreeEntries = nsTree.entrySet();
		  Iterator nsEntries = nsTreeEntries.iterator();
		  while(nsEntries.hasNext())
		  {
			  Map.Entry nsEntry = (Map.Entry) nsEntries.next();
			  
			  String nsKey = (String)nsEntry.getKey();
			  Hashtable objTree = (Hashtable)nsEntry.getValue();
			  
			  Set objEntrySet = objTree.entrySet();
			  Iterator objEntries = objEntrySet.iterator();
			  
			  while(objEntries.hasNext())
			  {
				  Map.Entry objEntry = (Map.Entry) objEntries.next();
				  
				  String objKey = (String)objEntry.getKey();
				  Hashtable revTree = (Hashtable)objEntry.getValue();
				  
				  Set revKeySet = revTree.keySet();
				  String[] revs = (String[]) revKeySet.toArray( vectorTemplate );
				  
				  for(int i=revs.length - 1; i>=0; i--)
				  { //do this so that they come out in the right order in the file
					  //this is important when we read the file back in.
					  fw.write("urn:lsid:");
					  fw.write( authKey );
					  fw.write(':');
					  fw.write( nsKey );
					  fw.write(':');
					  fw.write(objKey);
					  fw.write(':');
					  fw.write( (String) revs[i] );
					  fw.write('\n');
				  }
			  }
		  }
	  }
	  
	  fw.flush();
	  fw.close();
  }
  
  /**
   * read the tree file on disk and create a tree object from that.
   */
  public void readTree()
    throws IOException
  {
    FileReader fr = new FileReader(saveFile);
    StringBuffer sb = new StringBuffer();
    char[] c = new char[1024];
    int numread = fr.read(c, 0, 1024);
    while(numread != -1)
    {
      sb.append(c, 0, numread);
      numread = fr.read(c, 0, 1024);
    }
    String contents = sb.toString();
    StringTokenizer st = new StringTokenizer(contents);
    while(st.hasMoreTokens())
    {
      String token = st.nextToken();
      try
      {
        if(token != null)
        {
          addLSID(new KeplerLSID(token));
        }
      }
      catch(Exception e)
      {
        e.printStackTrace();
        throw new RuntimeException("A duplicate LSID was found in the " +
          "serialized LSID file " + saveFile.getAbsolutePath() + ".  This " +
          "shouldn't happen unless someone was mucking with the file. " +
          "You can try deleting the file and the cache and reloading " +
          "all system objects: " + e.getMessage());
      }
    }
  }

  /**
   * clears the tree and deletes the serial file.
   */
  public void clearTree()
  {
    authTree = new Hashtable();
    saveFile.delete();
  }

  /**
   * returns the next possible object number for an authority and namespace
   *
   * @param authority
   * @param namespace
   */
  private int getNextObject(String authority, String namespace)
    throws LSIDException
  {
    KeplerLSID lsid;
    Hashtable nsTree = (Hashtable)authTree.get(authority);
    if(nsTree != null)
    { //check if the authority exists
      Hashtable objTree = (Hashtable)nsTree.get(namespace);
      if(objTree != null)
      { //check if the namespace exists
        int highestObj = 0;
        Enumeration keys = objTree.keys();
        while(keys.hasMoreElements())
        {
          int key = (new Integer((String)keys.nextElement())).intValue();
          if(key > highestObj)
          {
            highestObj = key;
          }
        }

        return highestObj + 1;
      }
      else
      {
        return 1;
      }
    }
    else
    {
      return 1;
    }
  }
}
