/* Class representing a phylogenetic tree.
 *
 * Copyright (c) 2005 Natural Diversity Discovery Project.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE NATURAL DIVERSITY DISCOVERY PROJECT BE LIABLE 
 * TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 * CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND 
 * ITS DOCUMENTATION, EVEN IF THE NATURAL DIVERSITY DISCOVERY PROJECT
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE NATURAL DIVERSITY DISCOVERY PROJECT SPECIFICALLY DISCLAIMS ANY 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE NATURAL 
 * DIVERSITY DISCOVERY PROJECT HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

   @ProposedRating Red (tmcphillips@naturaldiversity.org)
   @AcceptedRating Red (tmcphillips@naturaldiversity.org) 
 */

package org.nddp.phylogeny;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.nddp.DomainObjectBase;
import org.nddp.exceptions.InconsistentDataException;
import org.nddp.exceptions.ParseException;
import org.nddp.util.Xml;

public class Tree extends DomainObjectBase {

	protected Tree() {
		super();
	}

	public Tree(Tree tree) {
	    	super();
	    	_id = tree._id;
		_characterMatrix = tree._characterMatrix;
	    _root = new TreeNode(this, tree._root, null);
	    _distancesDefined = tree._distancesDefined;
	    _characterStateChanges = 
	        (double[]) tree._characterStateChanges.clone();
	}	

    public Tree(String newickTreeString, 
            CharacterMatrix characterMatrix, Map expandedNameMap) 
        throws ParseException{

        if (newickTreeString == null) {
            throw new NullPointerException("null newickTreeString");
        }
        
        if (characterMatrix == null) {
            throw new NullPointerException("null characterMatrix");
        }
        
        _characterMatrix = characterMatrix;
        
        // OK if null
        _expandedNameMap = expandedNameMap;
        
        _characterStateChanges = new double[characterMatrix.characterCount()];

            // remove comment blocks and white space from tree file string
        newickTreeString = NEWICK_COMMENT_PATTERN.
            matcher(newickTreeString).replaceAll("");
            
        newickTreeString = NEWICK_WHITE_SPACE_PATTERN.
            matcher(newickTreeString).replaceAll("");
        
        newickTreeString = 
            newickTreeString.substring(1, newickTreeString.length() - 1);
        
        if (newickTreeString.indexOf(':') != -1) {
            _distancesDefined = true;
        }
        
            _root = new TreeNode(this, null, 0.0, newickTreeString);
    }

    ///////////////////////////////////////////////////////////////////
	////                         public methods                    //// 

	
	public final int characterCount() { 
	    return _characterMatrix.characterCount(); 
    };
	
    public CharacterMatrix characterMatrix() {
	    return _characterMatrix;
	}
	
	public final double characterStateChanges(int characterIndex) {
	    return _characterStateChanges[characterIndex];
	}

	public double consistencyIndex(int characterIndex) { 
	    return minimumStateChanges(characterIndex) / 
	    		characterStateChanges(characterIndex);
	}
	
	public double ensembleConsistencyIndex() {
	    
	    if (! _ensembleStatisticsCalculated) {
	        calculateEnsembleStatistics();
	    }
	    
	    return _ensembleConsistencyIndex;
	} 	
		
	public double ensembleMaximumPossibleChanges() {

	    if (! _ensembleStatisticsCalculated) {
	        calculateEnsembleStatistics();
	    }
	    
	    return _ensembleMaximumChanges;
	}
	
	public double ensembleMinimumStateChanges() {

	    if (! _ensembleStatisticsCalculated) {
	        calculateEnsembleStatistics();
	    }
	    
		return _ensembleMinimumChanges;
	}
	
	public double ensembleRetentionIndex() {
	    
	    if (! _ensembleStatisticsCalculated) {
	        calculateEnsembleStatistics();
	    }
	    
	   	return _ensembleRetentionIndex;
	}
	
 	public int equivalenceHashCode() {
	    
	    if (_equivalenceHashCode == 0 || ! isWriteLocked()) {
	        _equivalenceHashCode = _root.equivalenceHashCode();
	    }
	    
	    return _equivalenceHashCode;
	}
 	
	public boolean equivalentTo(Tree tree) {
	    return equivalenceHashCode() == tree.equivalenceHashCode();
	}
    
    public String id() {
	    return _id;
	}
    
	public final double maximumPossibleChanges(int characterIndex) {
		return _characterMatrix.maximumPossibleChanges(characterIndex);
	}
	
	public final double minimumStateChanges(int characterIndex) {
	    
	    return 
	    		_characterMatrix.character(characterIndex).minimumStateChanges();
	}
	
	public String newickString(boolean abbreviated, boolean hideLength) {
		return _root.toString(abbreviated, hideLength);
	}
	
	public int nodeCount() { 
	    return _nodeCount; 
	}
	
	public void print() { 
	    
	    _root.print("--"); 
	}
	
	public void printStatistics() {
	    
	    System.out.println("#  steps  min   max   ci    ri    rci");
	    for (int i = 0; i < characterCount(); i++ ) {
	        System.out.println(
                (i+1) + "   " + 
                characterStateChanges(i) + "   " + 
                minimumStateChanges(i) + "   " +
                maximumPossibleChanges(i) + "   " + 
                consistencyIndex(i) + "   " +
                retentionIndex(i) + "   " +
                rescaledConsistencyIndex(i)
            );
	    }
	    System.out.println("CI = " + ensembleConsistencyIndex());
	    System.out.println("RI = " + ensembleRetentionIndex());
	    System.out.println("RCI = " + rescaledEnsembleConsistencyIndex());
	    System.out.println("sumS = " + treeLength());
	    System.out.println("sumM = " + ensembleMinimumStateChanges());
	    System.out.println("sumG = " + ensembleMaximumPossibleChanges());
	}
	
	public static void removeEquivalentTrees(List treeList) {

	    for (int i = 0; i < treeList.size(); i++) {
	        
	        Tree treeOne = (Tree)treeList.get(i);
	        
	        for (int j = i + 1; j < treeList.size(); ) {
	        
	            Tree treeTwo = (Tree)treeList.get(j);
	            
	            if (treeOne.equivalentTo(treeTwo)) {
	                treeList.remove(j);
	            } else {
	                j++;
	            }
	        }
	    }   
	}
    public void reroot(String nodeName) throws InconsistentDataException {
        
        assertNotWriteLocked();
        
        TreeNode node = (TreeNode)_nameToNodeMap.get(nodeName);
            
            if (node == null) {
            throw new InconsistentDataException(
                "No node with name: " + nodeName);
        }
        
            if (node != _root) {
                _oldRoot = _root;
                node.reroot(null);
                _root = node;
            }       
    }
	
	public double rescaledConsistencyIndex(int characterIndex) {
	    
	    double ci = consistencyIndex(characterIndex);
	    double ri = retentionIndex(characterIndex);
	    
	    return ci * ri;
	}

	public double rescaledEnsembleConsistencyIndex() {
	    
	    if (! _ensembleStatisticsCalculated) {
	        calculateEnsembleStatistics();
	    }
	    
	    return _rescaledEnsembleConsistencyIndex;
	}
	
	public double retentionIndex(int characterIndex) {
	    
	    double g = maximumPossibleChanges(characterIndex);
	    double s = characterStateChanges(characterIndex);
	    double m = minimumStateChanges(characterIndex);
	    
	    	if ( g == s ) {
    			return 0;
	    	} else {	        
    			return (g - s) / (g - m);
	    	}
	}	
	
	public TreeNode root() {
	    return _root;
	}
    
    public void setCharacterChange(int characterIndex, double changes) {
        
        assertNotWriteLocked();

        _characterStateChanges[characterIndex] = changes;
        }
        
        public void setName(String name) {
        assertNotWriteLocked();

        _id = name;
    }
    

    public void setNodeStates(String[] nodeName, String[] states) 
        throws InconsistentDataException {
    
        assertNotWriteLocked();
    
        for (int i = 0; i < nodeName.length; i++ ) {
            
            String name = _expandName(nodeName[i]);
            TreeNode node = (TreeNode)_nameToNodeMap.get(name);
            
            if (node == null) {
                
                throw new InconsistentDataException(
                    "No node with name: " + name);
            }
            
            if (! node.isTerminalNode()) {
                node._setStates(states[i]);
            }
        }
    }   

    public void setParentNames(String[] childNameArray, 
            String[] parentNameArray) throws InconsistentDataException {

        assertNotWriteLocked();

        // make a local copy of the parent name array so that it 
            // can be modified
            parentNameArray = (String[]) parentNameArray.clone();
            
            // repeat scan of arrays of names until every node has a name
            while ( _nameToNodeMap.size() < _nodeCount) {
                
                // reset flag indicating whether a name was added on this
                // pass through the name arrays
                boolean nameAdded = false;
                
                // loop over indices of name arrays
            for (int i = 0; i < parentNameArray.length; i++ ) {
                
                // extract the parent and child names corresponding to 
                // the array index
                String parentName = _expandName(parentNameArray[i]);
                String childName = _expandName(childNameArray[i]);
                
                // check if parentName
                if ( parentName != null) {
                    if (! _nameToNodeMap.containsKey(parentName)) {
                        
                        TreeNode child = 
                            (TreeNode)_nameToNodeMap.get(childName);
                        
                        if ( child != null) {
                        
                            TreeNode parent = (TreeNode)child.parent();
                            
                            if (parent != null) {
                                parent._setName(parentName);
                            }
                            
                            parentNameArray[i] = null;
                            nameAdded = true;
                        }
                        }
                    }
                }
            
            if (! nameAdded) {
                System.out.println();
                throw new InconsistentDataException("Missing name");
            }
        }
    }   
    
    public void setWeightedCharacterChange(int characterIndex, 
            double weightedChanges) {
        
    assertNotWriteLocked();
        
    int weight = _characterMatrix.weight(characterIndex); 
        
        if ( weight == 0) {
                _characterStateChanges[characterIndex] = 0;
        } else {
            _characterStateChanges[characterIndex] = 
                    weightedChanges / weight;
        }
    }

    public String toString() { 
	    return newickString(true, true); 
    }
	
	public double treeLength() {

	    if (! _ensembleStatisticsCalculated) {
	        calculateEnsembleStatistics();
	    }
	    
		return _treeLength;
	}

	public String type() {
	    return "Tree";
	}

	public double weight(int characterIndex) {
	    return _characterMatrix.weight(characterIndex);
	}
	
	public String xmlAttributeString() {
	    return Xml.attribute("tree", newickString(true, true)); 
    }
	
	public String xmlContentString(Xml.Indentation indentation) {
	    return null;
	}
	
	///////////////////////////////////////////////////////////////////
	////                         protected variables               ////

	protected CharacterMatrix _characterMatrix;
	protected double[] _characterStateChanges;
	protected boolean _distancesDefined = false;
	protected String _id;
	protected int _nodeCount = 0;
	protected TreeNode _root;

    ///////////////////////////////////////////////////////////////////
    ////                         protected methods                 ////

        protected String _expandName(String name) {
            String expandedName;
            if ( _expandedNameMap != null && (expandedName = 
                (String)_expandedNameMap.get(name)) != null) {
                
            return expandedName;
            } else {
                return name;
            }
    }
    
    ///////////////////////////////////////////////////////////////////
	////                         private methods                   ////
	
	private void calculateEnsembleStatistics() {

	    _ensembleMaximumChanges = 0.0;
	    _ensembleMinimumChanges = 0.0;
	    _treeLength = 0.0;
	    
	    for (int i = 0; i < characterCount(); i++ ) {
	        	double weight = weight(i);  
	       	_ensembleMaximumChanges += weight * maximumPossibleChanges(i);
	       	_ensembleMinimumChanges += weight * minimumStateChanges(i);
	       	_treeLength += weight * characterStateChanges(i);
	    }
	    
	    _ensembleRetentionIndex = 
	        (_ensembleMaximumChanges - _treeLength) / 
	        (_ensembleMaximumChanges - _ensembleMinimumChanges);
	    _ensembleConsistencyIndex = _ensembleMinimumChanges/_treeLength;
	    _rescaledEnsembleConsistencyIndex = 
	        _ensembleRetentionIndex * _ensembleConsistencyIndex;
	    
	    _ensembleStatisticsCalculated = true;
	}
	
	///////////////////////////////////////////////////////////////////
	////                         private variables                 ////	
    
	private double _ensembleConsistencyIndex;
	private double _ensembleMaximumChanges;
	private double _ensembleMinimumChanges;
	private double _ensembleRetentionIndex;
	private boolean _ensembleStatisticsCalculated = false;	
	private int _equivalenceHashCode = 0;
    private Map _expandedNameMap;
    final Map _nameToNodeMap = new HashMap();    
    private static final Pattern NEWICK_COMMENT_PATTERN = 
        Pattern.compile("\\[.*?\\]", Pattern.DOTALL);    
    private static final Pattern NEWICK_WHITE_SPACE_PATTERN = 
        Pattern.compile("\\s*");
    TreeNode _oldRoot;
	private double _rescaledEnsembleConsistencyIndex;
	private double _treeLength;
}
