/* JUnit tests for CollectionIOPort class.
 *
 * Copyright (c) 2005 Natural Diversity Discovery Project.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE NATURAL DIVERSITY DISCOVERY PROJECT BE LIABLE 
 * TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 * CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND 
 * ITS DOCUMENTATION, EVEN IF THE NATURAL DIVERSITY DISCOVERY PROJECT
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE NATURAL DIVERSITY DISCOVERY PROJECT SPECIFICALLY DISCLAIMS ANY 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE NATURAL 
 * DIVERSITY DISCOVERY PROJECT HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

   @ProposedRating Red (tmcphillips@naturaldiversity.org)
   @AcceptedRating Red (tmcphillips@naturaldiversity.org) 
 */
 
package org.nddp;

import java.util.LinkedList;
import java.util.List;

import org.nddp.exceptions.CollectionException;
import org.nddp.tokens.ClosingDelimiterToken;
import org.nddp.tokens.DataToken;
import org.nddp.tokens.ExceptionToken;
import org.nddp.tokens.LoopTerminationToken;
import org.nddp.tokens.MetadataToken;
import org.nddp.tokens.OpeningDelimiterToken;
import ptolemy.actor.Director;
import ptolemy.actor.Manager;
import ptolemy.actor.NoTokenException;
import ptolemy.actor.Receiver;
import ptolemy.actor.TypedActor;
import ptolemy.actor.util.FunctionDependency;
import ptolemy.data.DoubleToken;
import ptolemy.data.IntToken;
import ptolemy.data.StringToken;
import ptolemy.data.Token;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.ComponentEntity;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.StringAttribute;
import junit.framework.TestCase;

/** 
 @author Timothy M. McPhillips
 */

public class TestCollectionIOPort extends TestCase {

	///////////////////////////////////////////////////////////////////
	////                         public methods                    //// 

    public void setUp() throws Exception {
        
        componentEntity = new ComponentEntity_Local();
        collection = new Collection();
    }
    
    public void testCreateOutputPort() throws Exception {

        CollectionIOPort port 
        		= CollectionIOPort.createOutputPort(componentEntity, "port name");
        
        assertEquals(componentEntity, port.getContainer());
        assertEquals("port name", port.getName());
        assertFalse(port.isInput());
        assertTrue(port.isOutput());
        assertEquals(BaseType.GENERAL, port.getType());
    }

    public void testCreateOutputPort_emptyName() throws Exception {

        try {
	        CollectionIOPort port 
	        		= CollectionIOPort.createOutputPort(componentEntity, "");
        
	        	fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            
        		assertEquals("Empty name not allowed.", 
        			expected.getMessage());
        }
    }
    
    public void testCreateOutputPort_nullContainer() throws Exception {

        try {
	        CollectionIOPort port 
    		    		= CollectionIOPort.createOutputPort(null, "port name");
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals("Null container not allowed.",
    		        expected.getMessage());
        }
    }
    
    public void testCreateOutputPort_nullName() throws Exception {

        try {
	        CollectionIOPort port 
    		    		= new CollectionIOPort(componentEntity, null);
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals("Null name not allowed.", 
        			expected.getMessage());
        }
    }
    
    public void testGet_DataToken() throws Exception {
        
        	CollectionIOPort_Local port = new CollectionIOPort_Local(
            	componentEntity, "port name", true, false);
        	Token stringToken = new StringToken("string");
        	DataToken dataToken = new DataToken(collection, stringToken);
        port.writeToInputQueue(dataToken);
        assertEquals(dataToken, port.get(false));
    }

    public void testGet_DataTokenUnwrapped() throws Exception {
        
        	CollectionIOPort_Local port = new CollectionIOPort_Local(
            	componentEntity, "port name", true, false);
        	
        	Token stringToken = new StringToken("string");
        	DataToken stringDataToken = new DataToken(collection, stringToken);
        	port.writeToInputQueue(stringDataToken);
        	
        	Token intToken = new IntToken(0);
        	DataToken intDataToken = new DataToken(collection, intToken);
        	port.writeToInputQueue(intDataToken);
        	
        	Token doubleToken = new DoubleToken(3.14149);
        	DataToken doubleDataToken = new DataToken(collection, doubleToken);
        	port.writeToInputQueue(doubleDataToken);
        	
        	assertEquals(stringToken, port.get(true));
        	assertEquals(intToken, port.get(true));
        	assertEquals(doubleToken, port.get(true));
    }

    public void testGet_DataTokenWrapped() throws Exception {
        
        	CollectionIOPort_Local port = new CollectionIOPort_Local(
            	componentEntity, "port name", true, false);
        	
        	Token stringToken = new StringToken("string");
        	DataToken stringDataToken = new DataToken(collection, stringToken);
        	port.writeToInputQueue(stringDataToken);
        	
        	Token intToken = new IntToken(0);
        	DataToken intDataToken = new DataToken(collection, intToken);
        	port.writeToInputQueue(intDataToken);
        	
        	Token doubleToken = new DoubleToken(3.14149);
        	DataToken doubleDataToken = new DataToken(collection, doubleToken);
        	port.writeToInputQueue(doubleDataToken);
        	
        	assertEquals(stringDataToken, port.get(false));
        	assertEquals(intDataToken, port.get(false));
        	assertEquals(doubleDataToken, port.get(false));
    }

    public void testGet_DelimiterTokens() throws Exception {
        
        	CollectionIOPort_Local port = new CollectionIOPort_Local(
            	componentEntity, "port name", true, false);
        	OpeningDelimiterToken openingDelimiterToken = 
        	    new OpeningDelimiterToken(collection);
        	ClosingDelimiterToken closingDelimiterToken = 
        	    new ClosingDelimiterToken(collection);
        	
        	port.writeToInputQueue(openingDelimiterToken);
        	port.writeToInputQueue(closingDelimiterToken);

        	assertEquals(null, port.get());
        	assertEquals(openingDelimiterToken, componentEntity.lastOpeningDelimiterToken);

        	assertEquals(null, port.get());
        	assertEquals(closingDelimiterToken, componentEntity.lastClosingDelimiterToken);
    }
    
    public void testGet_ExceptionToken() throws Exception {
        
        	CompositeEntity compositeEntity = new CompositeEntity();
       	CollectionActor collectionActor = 
            new CollectionActor(compositeEntity, "actor");
    		CollectionIOPort_Local port = new CollectionIOPort_Local(
            	componentEntity, "port name", true, false);
        	ExceptionToken exceptionToken = 
        	    new ExceptionToken(collection, collectionActor, 
				new CollectionException("error message"));

        	port.writeToInputQueue(exceptionToken);

        	assertEquals(null, port.get());
        	assertEquals(exceptionToken, componentEntity.lastExceptionToken);
    }    

    public void testGet_LoopTerminationToken() throws Exception {
        
        	CollectionIOPort_Local port = new CollectionIOPort_Local(
            	componentEntity, "port name", true, false);
        	LoopTerminationToken loopTerminationToken = 
        	    new LoopTerminationToken(collection);

        	port.writeToInputQueue(loopTerminationToken);

        	assertEquals(null, port.get());
        	assertEquals(loopTerminationToken, componentEntity.lastLoopTerminationToken);
    }
    
    public void testGet_MetadataToken() throws Exception {
        
        	CollectionIOPort_Local port = new CollectionIOPort_Local(
            	componentEntity, "port name", true, false);
        	MetadataToken metadataToken = 
        	    new MetadataToken(collection, "name", new StringToken("value"));

        	port.writeToInputQueue(metadataToken);

        	assertEquals(null, port.get());
        	assertEquals(metadataToken, componentEntity.lastMetadataToken);
    }

 	public void testGet_OutputPort() throws Exception {
        
        CollectionIOPort port = CollectionIOPort.createOutputPort(
                componentEntity, "port name");
        try {
            port.get();

            fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            assertEquals("Not an input port.",
                expected.getMessage());
        }
    }

    public void testGet_PrimitiveTokens() throws Exception {
        
        	CollectionIOPort_Local port = new CollectionIOPort_Local(
            	componentEntity, "port name", true, false);
        	Token stringToken = new StringToken("string");
        	port.writeToInputQueue(stringToken);
        	Token intToken = new IntToken(0);
        	port.writeToInputQueue(intToken);
        	Token doubleToken = new DoubleToken(3.14149);
        	port.writeToInputQueue(doubleToken);
        	assertEquals(stringToken, port.get(true));
        	assertEquals(intToken, port.get(true));
        	assertEquals(doubleToken, port.get(true));
    }

    public void testSend() throws Exception {

        	CollectionIOPort_Local port 
	    		= new CollectionIOPort_Local(componentEntity, "port name", 
    		        false, true);
		Token token = new StringToken("token");
        	assertEquals(null, port.lastBroadcastToken());
        	port.send(token);
        	assertEquals(token, port.lastBroadcastToken());
    }

    public void testSend_InputPort() throws Exception {
        
        	CollectionIOPort port 
	    		= CollectionIOPort.createInputPort(componentEntity, "port name");
		Token token = new StringToken("token");
		
        try {
            port.send(token);

            fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            assertEquals("Not an output port.",
                expected.getMessage());
        }
    }
    
    public void testSend_NullToken() throws Exception {
        
        	CollectionIOPort port 
	    		= CollectionIOPort.createOutputPort(componentEntity, "port name");
		
        try {
            port.send(null);
            
            fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            assertEquals("Null token not allowed.",
                expected.getMessage());
        }
    }

    public void testSetDirection() throws Exception {

        CollectionIOPort port = CollectionIOPort.createInputPort(
                componentEntity, "port name");
        StringAttribute attribute;
        
        port.setDirection("north");
        attribute = (StringAttribute) port.getAttribute("_cardinal");
        assertEquals("north", attribute.getExpression());

        port.setDirection("east");
        attribute = (StringAttribute) port.getAttribute("_cardinal");
        assertEquals("east", attribute.getExpression());

        port.setDirection("south");
        attribute = (StringAttribute) port.getAttribute("_cardinal");
        assertEquals("south", attribute.getExpression());

        port.setDirection("west");
        attribute = (StringAttribute) port.getAttribute("_cardinal");
        assertEquals("west", attribute.getExpression());
    }
    
    public void testSetDirection_BadDirection() throws Exception {

        CollectionIOPort port = CollectionIOPort.createInputPort(
                componentEntity, "port name");
        StringAttribute attribute;

        try {
	        port.setDirection("northwest");
	        
	        fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            
            assertEquals("Port direction must be one of " +
                "north, east, south and west.",
                expected.getMessage());
        }
    }

    public void testThreeArgumentConstructor_emptyName() throws Exception {

        CollectionIOPort.NondataHandler nondataHandler = new ComponentEntity_Local();

        try {
        CollectionIOPort port 
        		= CollectionIOPort.createInputPort(componentEntity, "",
        		        nondataHandler);
        
	        	fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            
        		assertEquals("Empty name not allowed.", 
        			expected.getMessage());
        }
    }

    public void testThreeArgumentConstructor_nullName() throws Exception {

        CollectionIOPort.NondataHandler nondataHandler = new ComponentEntity_Local();

        try {
	        CollectionIOPort port 
    		    		= CollectionIOPort.createInputPort(componentEntity, null,
        		        nondataHandler);
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals("Null name not allowed.", 
        			expected.getMessage());
        }
    }

    public void testThreeArgumentCreateInputPort() throws Exception {

        CollectionIOPort.NondataHandler nondataHandler = new ComponentEntity_Local();

        CollectionIOPort port 
        		= CollectionIOPort.createInputPort(componentEntity, "port name",
        		        nondataHandler);
        
        assertEquals(componentEntity, port.getContainer());
        assertEquals("port name", port.getName());
        assertTrue(port.isInput());
        assertFalse(port.isOutput());
        assertEquals(BaseType.GENERAL, port.getType());
        assertEquals(nondataHandler, port.nondataHandler());
    }  

    public void testThreeArgumentCreateInputPort_nullContainer() throws Exception {

        CollectionIOPort.NondataHandler nondataHandler = new ComponentEntity_Local();

        try {
	        CollectionIOPort port 
    		    		= CollectionIOPort.createInputPort(null, "port name",
	    		        nondataHandler);
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals("Null container not allowed.",
    		        expected.getMessage());
        }
    } 
    
    public void testThreeArgumentCreateInputPort_nullNondataHandler() throws Exception {

        try {
	        CollectionIOPort port 
    		    		= CollectionIOPort.createInputPort(componentEntity, "port name",
	    		        null);
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals("Null nondataHandler not allowed.",
    		        expected.getMessage());
        }
    }    

    public void testTwoArgumentConstructor() throws Exception {
      
        CollectionIOPort port 
        		= new CollectionIOPort(componentEntity, "port name");
        
        assertEquals(componentEntity, port.getContainer());
        assertEquals("port name", port.getName());
        assertTrue(port.isInput());
        assertFalse(port.isOutput());
        assertEquals(BaseType.GENERAL, port.getType());
        assertEquals(componentEntity, port.nondataHandler());
    }
    
    public void testTwoArgumentConstructor_emptyName() throws Exception {

        try {
	        CollectionIOPort port 
    		    		= new CollectionIOPort(componentEntity, "");
        
	        	fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            
        		assertEquals("Empty name not allowed.", 
        			expected.getMessage());
        }
    }
    
    public void testTwoArgumentConstructor_nullContainer() throws Exception {

        try {
	        CollectionIOPort port 
    		    		= new CollectionIOPort(null, "port name");
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals(null, 
        			expected.getMessage());
        }
    }
    
    public void testTwoArgumentConstructor_nullName() throws Exception {

        try {
	        CollectionIOPort port 
    		    		= new CollectionIOPort(componentEntity, null);
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals("Null name not allowed.", 
        			expected.getMessage());
        }
    }
    
    public void testTwoArgumentCreateInputPort() throws Exception {
      
        CollectionIOPort port 
        		= CollectionIOPort.createInputPort(componentEntity, "port name");
        
        assertEquals(componentEntity, port.getContainer());
        assertEquals("port name", port.getName());
        assertTrue(port.isInput());
        assertFalse(port.isOutput());
        assertEquals(BaseType.GENERAL, port.getType());
        assertEquals(componentEntity, port.nondataHandler());
    }
      
    public void testTwoArgumentCreateInputPort_emptyName() throws Exception {

        try {
	        CollectionIOPort port 
	        		= CollectionIOPort.createInputPort(componentEntity, "");
        
	        	fail("An IllegalActionException should have been thrown.");
        }
        catch(IllegalActionException expected) {
            
        		assertEquals("Empty name not allowed.", 
        			expected.getMessage());
        }
    }
    
    public void testTwoArgumentCreateInputPort_NotANondataHandler() throws Exception {

        ComponentEntity container = new ComponentEntity();
        
        try {
	        CollectionIOPort port 
    		    		= CollectionIOPort.createInputPort(container, "port name");
        
	        	fail("A ClassCastException should have been thrown.");
        }
        catch(ClassCastException expected) {
            
        		assertEquals("Container must implement NondataHandler interface.", 
        			expected.getMessage());
        }
    }

    public void testTwoArgumentCreateInputPort_nullContainer() throws Exception {

        try {
	        CollectionIOPort port 
    		    		= CollectionIOPort.createInputPort(null, "port name");
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals("Null container not allowed.",
    		        expected.getMessage());
        }
    }
    
    public void testTwoArgumentCreateInputPort_nullName() throws Exception {

        try {
	        CollectionIOPort port 
    		    		= new CollectionIOPort(componentEntity, null);
        
	        	fail("A NullPointerException should have been thrown.");
        }
        catch(NullPointerException expected) {
            
        		assertEquals("Null name not allowed.", 
        			expected.getMessage());
        }
    }

    ///////////////////////////////////////////////////////////////////
	////                         private variables                 ////

    private ComponentEntity_Local componentEntity;
	private Collection collection;
    	
	///////////////////////////////////////////////////////////////////
	////               private inner classes                     ////
	
    	private class CollectionIOPort_Local extends CollectionIOPort {

		public CollectionIOPort_Local(ComponentEntity container, String name, 
        		boolean isInput, boolean isOutput) throws IllegalActionException, 
        		NameDuplicationException {
     
	    		super(container, name, isInput, isOutput);	
    		}

		///////////////////////////////////////////////////////////////////
		////          public methods for CollectionIOPort_Local        //// 

		public void broadcast(Token token) throws IllegalActionException {
	        lastBroadcastToken = token;
	        super.broadcast(token);
	    }

	    public Token get(int channel) throws NoTokenException, IllegalActionException {
	        
	        return (Token)tokenBuffer.remove(0);
	    }
	    
	    public Token lastBroadcastToken() {
	        return lastBroadcastToken;
	    }

		///////////////////////////////////////////////////////////////////
		////         private methods for CollectionIOPort_Local        //// 

	    private void writeToInputQueue(Token token) {
	    		tokenBuffer.add(token);
	    }
	    
		///////////////////////////////////////////////////////////////////
		////       private variables for CollectionIOPort_Local        //// 

	    private Token lastBroadcastToken = null;
    		private List tokenBuffer = new LinkedList();
    }
    	
	private class ComponentEntity_Local extends ComponentEntity 
    		implements CollectionIOPort.NondataHandler, TypedActor {

		///////////////////////////////////////////////////////////////////
		////          public methods for ComponentEntity_Local         //// 

	    public void fire() throws IllegalActionException {
        }

        public Director getDirector() {
            return null;
        }

        public Director getExecutiveDirector() {
            return null;
        }

        public FunctionDependency getFunctionDependency() {
            return null;
        }

        public Manager getManager() {
            return null;
        }

        public void handleClosingDelimiter(ClosingDelimiterToken token) throws IllegalActionException {
			lastClosingDelimiterToken = token;
        }

        public void handleExceptionToken(ExceptionToken token) throws IllegalActionException {
			lastExceptionToken = token;
        }

        public void handleLoopTerminationToken(LoopTerminationToken token) throws IllegalActionException {
			lastLoopTerminationToken = token;
        }

        public void handleMetadataToken(MetadataToken token) throws IllegalActionException {
			lastMetadataToken = token;
        }

        public void handleOpeningDelimiter(OpeningDelimiterToken token) throws IllegalActionException {
			lastOpeningDelimiterToken = token;
        }

        public void initialize() throws IllegalActionException {
        }

        public List inputPortList() {
            return null;
        }

        public int iterate(int count) throws IllegalActionException {
            return 0;
        }

        public Receiver newReceiver() throws IllegalActionException {
            return null;
        }

        public List outputPortList() {
            return null;
        }

        public boolean postfire() throws IllegalActionException {
            return false;
        }

        public boolean prefire() throws IllegalActionException {
            return false;
        }

        public void preinitialize() throws IllegalActionException {
        }

        public void stop() {
        }

        public void stopFire() {
        }

        public void terminate() {
        }

        public List typeConstraintList() throws IllegalActionException {
            return null;
        }

        public void wrapup() throws IllegalActionException {
        }
        
		///////////////////////////////////////////////////////////////////
		////          private variables for ComponentEntity_Local      ////
	
        private ClosingDelimiterToken lastClosingDelimiterToken = null;
        private ExceptionToken lastExceptionToken = null;
        private LoopTerminationToken lastLoopTerminationToken = null;
        private MetadataToken lastMetadataToken = null;
        private OpeningDelimiterToken lastOpeningDelimiterToken = null;
    }
}
