/* Command line actor generates and executes a command.
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 *                                       PT_COPYRIGHT_VERSION_2
 *                                       COPYRIGHTENDKEY
 */


package org.sdm.spa;

// Ptolemy packages
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Iterator;
import java.util.List;

import ptolemy.actor.IOPort;
import ptolemy.actor.TypedAtomicActor;
import ptolemy.actor.TypedIOPort;
import ptolemy.actor.parameters.FilePortParameter;
import ptolemy.actor.parameters.PortParameter;
import ptolemy.data.ArrayToken;
import ptolemy.data.BooleanToken;
import ptolemy.data.RecordToken;
import ptolemy.data.StringToken;
import ptolemy.data.expr.FileParameter;
import ptolemy.data.expr.Parameter;
import ptolemy.data.expr.SingletonParameter;
import ptolemy.data.type.BaseType;
import ptolemy.data.type.Type;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.Attribute;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.InternalErrorException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Nameable;
//////////////////////////////////////////////////////////////////////////
//// CommandLineExec
/**
   Given a command string, the <I>CommandLineExec</I> actor executes it using the
   java Runtime class.
   <p>
   <b>SUPPORTED COMMAND TYPES:</b>
   <ul><li>command
   <li>command < infile > outfile
   <li>command > outfile
   <li>command < infile
   <li>command [arg1..argn] > outfile
   <li>command [arg1..argn]
   <li>command [arg1..argn] < infile > outfile
   <li>command1 | command 2 <I>(<B>Warning</B>: This type of commands doesn't give the output of all the commands. Instead it outputs only the result of the last one.)</I>
   </ul><p>
   <b>Example commands:</b>
   <ul><li>C:/Program Files/Internet Explorer/IEXPLORE.EXE
           <br>To generate this command, just double click on the actor and type this in the <i>command</i> parameter field.
   <li>C:/cygwin/bin/perl.exe c:/project/kepler/test/workflows/example.pl > c:/project/kepler/test/workflows/example.out
   <li>C:/cygwin/bin/dir.exe > dirTemp.txt
   </ul>
   @author Ilkay Altintas, Christopher Hylands Brooks, Bilsay Yildirim, Contributor: Edward A. Lee
   @version $Id: CommandLineExec.java,v 1.3 2005/11/01 20:39:14 ruland Exp $

   @category.name external execution
   @category.name local
 */

public class CommandLineExec
    extends TypedAtomicActor {

  /** Construct a CommandLine actor with the given container and name.
   *  @param container The container.
   *  @param name The name of this actor.
   *  @exception IllegalActionException If the actor cannot be contained
   *   by the proposed container.
   *  @exception NameDuplicationException If the container already has an
   *   actor with this name.
   */
  public CommandLineExec(CompositeEntity container, String name) throws
      IllegalActionException, NameDuplicationException {
    super(container, name);

    // Uncomment the next line to see debugging statements
    // addDebugListener(new ptolemy.kernel.util.StreamListener());



    // Construct parameters.
    command = new PortParameter(this, "command");

    // command.setExpression("Please type your command here...");
    outputFile = new FilePortParameter(this, "outputFile");




    environment = new Parameter(this, "environment");
        directory = new FileParameter(this, "directory");
     String[] labels = new String[] {
            "name",
            "value"
        };
        Type[] values = new Type[] {
            BaseType.STRING,
            BaseType.STRING
        };

        // An array of records {{name = "", value = ""}}
       // environment.setTypeEquals(new ArrayType(new RecordType(labels, values)));

        // Array with an empty name and value means
        // default environment of the calling process.



        environment.setExpression("{{name = \"\", value = \"\"}}");
           directory.setExpression("$CWD");



    // Construct input ports.
    arguments = new TypedIOPort(this, "arguments", true, false);
    inputStream = new TypedIOPort(this, "inputStream", true, false);
    arguments.setMultiport(true);
    infileHandle = new TypedIOPort(this, "infileHandle", true, false);
    infileHandle.setTypeEquals(BaseType.STRING);
    trigger = new TypedIOPort(this, "trigger", true, false);
    new Attribute(arguments, "_showName");
    new Attribute(inputStream, "_showName");
    new Attribute(infileHandle, "_showName");
    new Attribute(command, "_showName");
    //Attribute hide = new SingletonAttribute(trigger, "_hide");
    //hide.setPersistent(false);
    hide = new SingletonParameter(trigger, "_hide");  //DFH
    hide.setToken(BooleanToken.TRUE);                 //DFH
    // Construct output ports.
    outfileHandle = new TypedIOPort(this, "outfileHandle", false, true);
    outfileHandle.setTypeEquals(BaseType.STRING);
     exitCode = new TypedIOPort(this, "exitCode", false, true);
        exitCode.setTypeEquals(BaseType.STRING);

    output = new TypedIOPort(this, "output", false, true);
    output.setTypeEquals(BaseType.STRING);

      new Attribute(exitCode, "_showName");
    new Attribute(output, "_showName");
    new Attribute(outfileHandle, "_showName");
      new Attribute(outputFile, "_showName");

    // Set Flags.
    outputLineByLine = new Parameter(this, "outputLineByLine", new BooleanToken(false));
    outputLineByLine.setTypeEquals(BaseType.BOOLEAN);
    waitForProcess = new Parameter(this, "waitForProcess", new BooleanToken(false));
    waitForProcess.setTypeEquals(BaseType.BOOLEAN);
    hasTrigger = new Parameter(this, "hasTrigger", new BooleanToken(false));
    hasTrigger.setTypeEquals(BaseType.BOOLEAN);

    _attachText("_iconDescription",
                "<svg>\n"
                + "<rect x=\"0\" y=\"0\" "
                + "width=\"60\" height=\"30\" "
                + "style=\"fill:white\"/>\n"
                + "<text x=\"20\" y=\"25\" "
                + "style=\"font-size:30; fill:blue; font-family:SansSerif\">"
                + "$</text>\n"
                + "</svg>\n");

  } // constructor

////////////////// Public ports and parameters ///////////////////////


  public SingletonParameter hide;
  /** @entity.description The command to execute.
   * <BR> <I>FIX ME: </I>The style of the command will be noted here.
   */
  public PortParameter command;

  /** @entity.description Needs to be filled in if the user wants the command
   *  to output to a file.
   */
  public FilePortParameter outputFile;

  /** @entity.description The arguments to the command. Implemented as a
   * multi/input port to support more than one argument. It concatanates the
   * inputs in all the channels.
   * <BR> <I>If there is an input file, this port can be empty.</I>
   */
  public TypedIOPort arguments;

   /** @entity.description Strings to pass to the standard input of the subprocess.
    *  This port is an input port of type String.
    */
  public TypedIOPort inputStream;

  /** @entity.description This is an optional port. Used if the file accepts
   * an input file instead of a list of arguments.
   */
  public TypedIOPort infileHandle;

  /** @entity.description The trigger port.  The type of this port is
   * undeclared, meaning that it will resolve to any data type.
   */
  public TypedIOPort trigger;

  /**
   * @entity.description A string that forwards the outputFile parameter
   * if it exists.
   */
  public TypedIOPort outfileHandle;
  /**
   * @entity.description The result stream of the command.
   */
  public TypedIOPort output;

  /**
   * @entity.description Exit code will be 1 if the command executes successfully.
   */
  public TypedIOPort exitCode;

  /**
   * @entity.description If selected, broadcasts the output of the command line by line.
   *
   */
  public Parameter outputLineByLine;

  /** @entity.description Unhide the trigger port when this parameter
   *  is true. This Parameter is type of boolean.
   *  <BR><b>NOTE: </b>in fact, user can use the port configuration window to hide or unhide
   *  a port. This paremeter is here to provide a more intuitive interface for
   *  this actor.
   */
  public Parameter hasTrigger;

  /*@entity.description The environment to execute the command.
  **/
  public Parameter environment;
  /*@entity.description The directory to execute the command
   **/
  public FileParameter directory;
  /*@entity.description This allows parameter to allow the process to block the threads until it exits.
   *
   **/
  public Parameter waitForProcess;




  ///////////////////////////////////////////////////////////////////
  ////                        public methods                     ////

  /** If the specified attribute is <i>showTriggerPort</i>, then get
   *  the value of it and re-render the trigger port. If it is true,
   *  show the trigger port; if it is false, hide the trigger port.
   *  @param attribute The attribute that has changed.
   *  @exception IllegalActionException.
   */
  public void attributeChanged(Attribute attribute)
          throws IllegalActionException {
              if (attribute == hasTrigger) {
                _triggerFlag = ( (BooleanToken) hasTrigger.getToken()).booleanValue();
                _debug("<TRIGGER_FLAG>" + _triggerFlag + "</TRIGGER_FLAG>");
                if (_triggerFlag) {
                  try {
                    trigger.setContainer(this);
                    //new Attribute(trigger, "_showName");
                    hide.setToken(BooleanToken.FALSE);   //DFH
                  } catch (NameDuplicationException ndex) {
                    _debug("111: " + ndex.getMessage());
                  }
                }else {
                  List inPortList = this.inputPortList();
                  Iterator ports = inPortList.iterator();
                  while (ports.hasNext()) {
                    IOPort p = (IOPort) ports.next();
                    if (p.isInput()) {
                      try {
                        if (p.getName().equals("trigger")) {
                          //new Attribute(trigger, "_hideName");  //DFH
                      //    p.setContainer(null);              //DFH
                           hide.setToken(BooleanToken.TRUE);   //DFH
                        }
                      } catch (Exception e) {
                          throw new IllegalActionException(this, e.getMessage());
                      }
                    }
                  }//while
                }//else
      } else {
          super.attributeChanged(attribute);
      }
  }
  /** ...
   *  Send the exitCode and outputFileHandle(optional) to the result
   *  port.
   *
   *  @exception IllegalActionException If there is no director.
   */
  public void fire() throws IllegalActionException {

    _lineFlag = ( (BooleanToken) outputLineByLine.getToken()).booleanValue();
    _debug("<TRIGGER_FLAG>" + _lineFlag + "</TRIGGER_FLAG>");

    _waitForProcessFlag = ( (BooleanToken) waitForProcess.getToken()).booleanValue();




    //Get Directory
     directoryAsFile = directory.asFile();
     if (_debugging) {
            _debug("About to exec \""
                      + ((StringToken) command.getToken()).stringValue() + "\""
                      + "\n in \"" + directoryAsFile + "\"\n with environment:");
        }
    // Process the environment parameter.
            ArrayToken environmentTokens = (ArrayToken) environment.getToken();

            if (_debugging) {
                _debug("environmentTokens: " + environmentTokens);
            }

        environmentArray = null;

            if (environmentTokens.length() >= 1) {
                environmentArray = new String[environmentTokens.length()];

                for (int i = 0; i < environmentTokens.length(); i++) {
                    StringToken nameToken = (StringToken) (((RecordToken) environmentTokens
                                                                   .getElement(i)).get("name"));
                    StringToken valueToken = (StringToken) (((RecordToken) environmentTokens
                                                                    .getElement(i)).get("value"));
                    environmentArray[i] = nameToken.stringValue() + "="
                        + valueToken.stringValue();

                    if (_debugging) {
                        _debug("  " + i + ". \"" + environmentArray[i] + "\"");
                    }

                    if ((i == 0) && (environmentTokens.length() == 1)
                            && environmentArray[0].equals("=")) {
                        if (_debugging) {
                            _debug("There is only one element, "
                                    + "it is a string of length 0,\n so we "
                                    + "pass Runtime.exec() an null "
                                    + "environment so that we use\n "
                                    + "the default environment");
                        }

                        environmentArray = null;
                    }
                }
            }


     // Get the main command from the command parameter.
    // _commandStr = ((StringToken)command.getToken()).stringValue();
    command.update();

    if (((StringToken) command.getToken()).stringValue() != null) {
        String _commandStr = ((StringToken) command.getToken()).stringValue();


    //}

    //simply consume the trigger token if there is some.
    if (_triggerFlag) {
        if (trigger.getWidth() > 0) {
            if (!((BooleanToken) trigger.get(0)).equals(null)) {
                trigger.get(0);
                _debug("consume the token at the trigger port.");
            }
        }
    }

    /* Consume the input file token if there's one.
     */


    String value = null;
    try {
         if ((infileHandle.numberOfSources() > 0)) {
             if( ((StringToken)infileHandle.get(0)).stringValue() !=""){

           value = ((StringToken) infileHandle.get(0)).stringValue();

        _debug("infileHandle(i) = " + value);
        if(value.startsWith("file:/"))
        value=value.substring(7);
        //else{
    //	value = value.substring(1, value.length() - 1);
       // }

        _commandStr += " < " + value;

          }
      }
    }
    catch (Exception ex) {
      _debug("Input file is null.");
    }

    /* The arguments can only be accepted if there's no input file. So the
     * "value" of the infile handle is checked here and arguments are consumed
     * only if it is null.
     */
//    if (value == null) {
          /* Create the argument string. Consume data in all the channels an combine
       * them.
       */


      String argString = "";
      value = new String("");
      int i = 0;

      int width = arguments.getWidth();

      for (i = 0; i < width; i++) {
       if ((arguments.numberOfSources() > 0)) {
           if(((StringToken)arguments.get(i)).stringValue() !="")
           {

           value = ((StringToken) arguments.get(i)).stringValue();

          _debug("arguments(i) = " + value);
        //  value = value.substring(1, value.length());

          while (value.indexOf("\\\"") != -1) {
           int ind = value.indexOf("\\\"");
           value = value.substring(0,ind) + value.substring(ind+1, value.length());
              _debug(value);
         }
          argString += value + " ";
          _debug("argString = " + argString);
        }
        }

      }
      _commandStr += " " + argString;

//    }

    int commandCount = getSystemProps();
    outputFile.update();
     /* Get the output file path if there's one and add it to the command string.
     */
    if (((StringToken)outputFile.getToken()).stringValue() =="") {
      _debug("Output file is null.");
        outfileHandle.broadcast(new StringToken(""));
    }
    else {
      String outFilePath = outputFile.stringValue();

      if (outFilePath.startsWith("file:///")) {
        if (_charsToSkip == 6) {
          _charsToSkip = 8;
        }
        else if (_charsToSkip == 5) {
          _charsToSkip = 7;
        }
      }
      if (outFilePath.startsWith("file:/")) {
          outFilePath = outputFile.stringValue().substring(_charsToSkip);
          }
      _commandStr += " > " + outFilePath;

          if(!outFilePath.equals(""))
      outfileHandle.broadcast(new StringToken(outFilePath));
    }

    _commandArr[commandCount] = _commandStr;

    _debug("<COMMAND>" + _commandArr[commandCount] + "</COMMAND>");
    // EXECUTION OF THE GENERATED COMMAND.
    _debug("Executing the command...");




    try {
           _stopFireRequested = false;


      //Execute the command
      Runtime rt = Runtime.getRuntime();
      proc = rt.exec(_commandArr,environmentArray,
                    directoryAsFile);




           //StringBuffer errBuffer = new StringBuffer();
        InputStream errStream = proc.getErrorStream();
         _errorGobbler = new _StreamReaderThread(errStream,
                    "Exec Stderr Gobbler-" + _streamReaderThreadCount++, this);

       // StringBuffer inBuffer = new StringBuffer();
        InputStream inStream = proc.getInputStream();
        _outputGobbler = new _StreamReaderThread(inStream,
                    "Exec Stdout Gobbler-" + _streamReaderThreadCount++, this);

        _errorGobbler.start();
        _outputGobbler.start();

            if (_streamReaderThreadCount > 1000) {
                // Avoid overflow in the thread count.
                _streamReaderThreadCount = 0;
            }

                   String errorString = _errorGobbler.getAndReset();
        String outputString = _outputGobbler.getAndReset();
         if (_debugging) {
            _debug("Exec: Error: '" + errorString + "'");
            _debug("Exec: Output: '" + outputString + "'");
        }

        // We could have a parameter that if it was set
        // we would throw an exception if there was any error data.
           if(!procDone(proc))
          {

            String line=null;
            if ((inputStream.numberOfSources() > 0) ) {
                if(((StringToken) inputStream.get(0)).stringValue() != null)
                {

            if ((line = ((StringToken) inputStream.get(0)).stringValue()) != null) {
                if (_debugging) {
                    _debug("CommandLine: Input: '" + line + "'");
                }
                 OutputStream oStream=proc.getOutputStream();

                 OutputStreamWriter inputStreamWriter = new OutputStreamWriter(proc.getOutputStream());
                _inputBufferedWriter = new BufferedWriter(inputStreamWriter);
                if (_inputBufferedWriter != null) {
                    try {
                        _inputBufferedWriter.write(line);
                        _inputBufferedWriter.flush();
                    } catch (IOException ex) {
                        throw new IllegalActionException(this, ex,
                                "Problem writing input '" + line + "'");
                    }
                }
                }
            }
            }
        }


        //Send error and output to ports


            exitCode.send(0, new StringToken(errorString));

            output.send(0, new StringToken(outputString));



        //Wait for the Process to finish if it is indicated
        if(_waitForProcessFlag)
        {

           try {

            // The next line waits for the subprocess to finish.
            int processReturnCode = proc.waitFor();

            if (processReturnCode != 0) {
                // We could have a parameter that would enable
                // or disable this.

                throw new IllegalActionException(this,
                        "Executing command \""
                        + ((StringToken) command.getToken()).stringValue()
                        + "\" returned a non-zero return value of "
                        + processReturnCode);
            }
        }
         catch (InterruptedException interrupted) {
            throw new InternalErrorException(this, interrupted,
                    "_process.waitFor() was interrupted");
        }

        }

     /*
   if (proc != null) {
                // Note that we assume that _process is null upon entry
                // to this method, but we check again here just to be sure.
                _terminateProcess();
            }*/
          //  br.close();
           // _inputStream.close();

    }
    catch (Exception ex) {
      _debug("<EXCEPTION> An exception occured when executing the command. " +
             " \n\t Exception: " + ex + "\n</EXCEPTION>");
    }
    }
  } //end-of-fire

//Check if the process is done or not
private boolean procDone(Process p) {
        try {
            int v = p.exitValue();
            return true;
        }
        catch(IllegalThreadStateException e) {
            return false;
        }
    }
 //Get system properties and set the command array according to it
  public int getSystemProps() throws IllegalActionException {
    // Get OS name
    String osName = System.getProperty("os.name");
     //	System.out.println(((StringToken) shell.getToken()).stringValue().toString());
  // String sh="sh";
        _debug("<OS>" + osName + "</OS>");

    if (osName.equals("Windows NT") || osName.equals("Windows XP") ||
        osName.equals("Windows 2000")) {

      _commandArr[0] = "cmd.exe";
      _commandArr[1] = "/C";
      _charsToSkip = 6;
      /*	}
          else{

      _commandArr[0] ="C:'\''cygwin'\''cywgin.bat";
      _commandArr[1] = "-c";
      _charsToSkip = 6;
          }*/
      return 2;
    }
    else if (osName.equals("Windows 95")) {
      _commandArr[0] = "command.com";
      _commandArr[1] = "/C";
      _charsToSkip = 6;
      return 2;
    }
/*    else if (osName.equals("Linux")) {
      _commandArr[0] = "/bin/sh";
      _commandArr[1] = "-c";
      _charsToSkip = 5;
      return 2;
    }
    else if (osName.equals("Mac OS X")) {
      _commandArr[0] = "/bin/sh";
      _commandArr[1] = "-c";
      _charsToSkip = 5;
      return 2;
    }*/
    else {
      _commandArr[0] = "/bin/sh";
      _commandArr[1] = "-c";
      _charsToSkip = 5;
      return 2;
/*      return 0;*/
    }
  } //end-of-getSystemProps
      /** Override the base class to stop waiting for input data.
     */
    public void stopFire() {
        // NOTE: This method used to be synchronized, as
        // was the fire() method, but this caused deadlocks.  EAL
        super.stopFire();
        _stopFireRequested = true;

        try {
            _terminateProcess();
        } catch (IllegalActionException ex) {
            throw new InternalErrorException(ex);
        }
    }
        private void _terminateProcess() throws IllegalActionException {
        if (proc != null) {
            proc.destroy();
            proc = null;
        }
    }
      /** Terminate the subprocess.
     *  This method is invoked exactly once per execution
     *  of an application.  None of the other action methods should be
     *  be invoked after it.
     *  @exception IllegalActionException Not thrown in this base class.
     */
    public void wrapup() throws IllegalActionException {
        _terminateProcess();
    }


    ///////////////////////////////////////////////////////////////////
    ////                         inner classes                     ////
    // Private class that reads a stream in a thread and updates the
    // stringBuffer.
    private class _StreamReaderThread extends Thread {
        /** Create a _StreamReaderThread.
         *  @param inputStream The stream to read from.
         *  @param name The name of this StreamReaderThread,
         *  which is useful for debugging.
         *  @param actor The parent actor of this thread, which
         *  is used in error messages.
         */
        _StreamReaderThread(InputStream inputStream, String name, Nameable actor) {
            super(name);
            _inputStream = inputStream;
            _inputStreamReader = new InputStreamReader(_inputStream);
            _actor = actor;
            _stringBuffer = new StringBuffer();
        }



        /** Read any remaining data in the input stream and return the
         *  data read thus far.  Calling this method resets the
         *  cache of data read thus far.
         */
        public String getAndReset() {

            if (_debugging) {
                try {
                    _debug("getAndReset: Gobbler '" + getName() + "' Ready: "
                            + _inputStreamReader.ready() + " Available: "
                            + _inputStream.available());
                } catch (Exception ex) {
                    throw new InternalErrorException(ex);
                }
            }

            try {
                // Read any remaining data.
                _read();
            } catch (Throwable throwable) {
                if (_debugging) {
                    _debug("WARNING: getAndReset(): _read() threw an "
                            + "exception, which we are ignoring.\n"
                            + throwable.getMessage());
                }
                throwable.printStackTrace();
            }

            String results = _stringBuffer.toString();
            _stringBuffer = new StringBuffer();

        /*    try {
                br.close();
               / _inputStream.close();
                _inputStreamReaderClosed = true;
            } catch (Exception ex) {
                throw new InternalErrorException(null, ex,
                        getName() + " failed to close.");
            }*/

            return results;
        }

        /** Read lines from the inputStream and append them to the
         *  stringBuffer.
         */
        public synchronized void run() {
            if (!_inputStreamReaderClosed) {
                _read();
            }
        }

        // Read from the stream until we get to the end of the stream
        // This is synchronized so that it is not called simultaneously
        // from run() and getAndReset().
        private synchronized void _read() {
            // We read the data as a char[] instead of using readline()
            // so that we can get strings that do not end in end of
            // line chars.
            char[] chars = new char[80];
            int length; // Number of characters read.

            try {
                // Oddly, InputStreamReader.read() will return -1
                // if there is no data present, but the string can still
                // read.


            InputStreamReader isr = new InputStreamReader(_inputStream);
             br = new BufferedReader(isr);
            String line=null;

                while ((line = br.readLine()) != null && !_stopFireRequested) {
                             if(line.length()!=0)
                             {

                    if (_debugging) {
                        // Note that ready might be false here since
                        // we already read the data.
                        _debug("_read(): Gobbler '" + getName() + "' Ready: "
                                + line + " Value: '"
                                + line + "'");
                    }

                    _stringBuffer.append(line + "\n");
              }

           }
           } catch (IOException ex) {
                    throw new InternalErrorException(_actor, ex,
                        getName() + ": Failed while reading from " + _inputStream);

                    }

            }


        // The actor associated with this stream reader.
        private Nameable _actor;

        // Stream from which to read.
        private InputStreamReader _inputStreamReader;

        // Indicator that the stream has been closed.
        private boolean _inputStreamReaderClosed = false;


        // StringBuffer to update.
        private StringBuffer _stringBuffer;

        //BufferReader
        private  BufferedReader br;
    }



  //////////////////////////////////////////////////////////////////////
  ////                         private variables                    ////

  // StreamReader with which we read stderr.
    private _StreamReaderThread _errorGobbler;

  // Stream from which to read.
  private InputStream _inputStream;
  // StreamReader with which we read stdout.
  private _StreamReaderThread _outputGobbler;
  // The subprocess gets its input from this BufferedWriter.
  private BufferedWriter _inputBufferedWriter;
  // The combined command to execute.
  private String _commandStr = "";
  private String _commandArr[] = new String[3];
  private boolean _lineFlag = false;
  private boolean _waitForProcessFlag = false;
  private String[] environmentArray;
  private File directoryAsFile;
  private boolean environmentChanged = false;
  private boolean _triggerFlag = false;
  // Indicator that stopFire() has been called.
  private boolean _stopFireRequested = false;
  private Process proc;
  private int _charsToSkip = 6;
  // Instance count of output and error threads, used for debugging.
  // When the value is greater than 1000, we reset it to 0.
  private static int _streamReaderThreadCount = 0;
} // end-of-class-CommandLine
